"""
These tasks serves to 1) illustrate what types of tasks *can* be done with Norse
and 2) how to use Norse for specific tasks. Please note that some tasks require
additional dependencies, like OpenAI gym for the cartpole task, which is not included in
vanilla Norse.

Please refer to the page on :doc:`/tasks` for thorough explanations on how to execute and run the
experiments/tasks.
"""



r"""
In this task, we train a spiking convolutional network to learn the
MNIST digit recognition task.

This version uses the `PyTorch Lightning <https://pytorchlightning.ai/>`_ library
to reduce the amount of boilerplate code around logging, checkpointing, training, etc.
"""
from argparse import ArgumentParser

import torch
import torch.utils.data
import torchvision

import pytorch_lightning as pl

import norse.torch as norse


class LIFConvNet(pl.LightningModule):
    def __init__(
        self,
        input_features,
        seq_length,
        learning_rate,
        model="super",
        optimizer="adam",
        input_scale=1.0,
        only_output=False,
        only_first_spike=False,
    ):
        super(LIFConvNet, self).__init__()
        if only_first_spike:
            self.encoder = norse.SpikeLatencyLIFEncoder(seq_length=seq_length)
        else:
            self.encoder = norse.ConstantCurrentLIFEncoder(seq_length=seq_length)
        self.input_features = input_features
        self.input_scale = input_scale
        self.learning_rate = learning_rate
        self.only_output = only_output
        self.optimizer = optimizer
        self.rsnn = norse.ConvNet4(method=model)
        self.seq_length = seq_length

    def forward(self, x):
        batch_size = x.shape[0]
        x = self.encoder(x.view(-1, self.input_features))
        x = x.reshape(self.seq_length, batch_size, 1, 28, 28)
        voltages = self.rsnn(x)
        m, _ = torch.max(voltages, 0)
        return torch.nn.functional.log_softmax(m, dim=1)

    def training_step(self, batch, batch_idx):
        x, y = batch
        out = self(x)
        loss = torch.nn.functional.nll_loss(out, y)
        return loss

    def configure_optimizers(self):
        if self.optimizer == "adam":
            opt = torch.optim.Adam
        else:
            opt = torch.optim.SGD

        if self.only_output:
            return opt(
                self.rsnn.out.parameters(), lr=self.learning_rate, weight_decay=1e-6
            )

        return opt(self.rsnn.parameters(), lr=self.learning_rate, weight_decay=1e-6)


def main(args):
    # First we create and transform the dataset
    data_transform = torchvision.transforms.Compose(
        [
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.1307,), (0.3081,)),
        ]
    )
    train_loader = torch.utils.data.DataLoader(
        torchvision.datasets.MNIST(
            root=".",
            train=True,
            download=True,
            transform=data_transform,
        ),
        batch_size=args.batch_size,
        shuffle=True,
    )
    test_loader = torch.utils.data.DataLoader(
        torchvision.datasets.MNIST(
            root=".",
            train=False,
            transform=data_transform,
        ),
        batch_size=args.batch_size,
    )

    # Second, we create the PyTorch Lightning module
    model = LIFConvNet(
        28 * 28,  # Standard MNIST size
        seq_length=args.seq_length,
        learning_rate=args.learning_rate,
        model=args.model,
        optimizer=args.optimizer,
        only_output=args.train_only_output,
        only_first_spike=args.only_first_spike,
    )

    # Finally, we define the trainer and fit the model
    trainer = pl.Trainer.from_argparse_args(args)
    trainer.fit(model, train_loader)
    trainer.test(test_dataloader=test_loader)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser = pl.Trainer.add_argparse_args(parser)
    parser.set_defaults(
        max_epochs=1000, auto_select_gpus=True, progress_bar_refresh_rate=1
    )
    parser.add_argument(
        "--batch_size", default=32, type=int, help="Number of examples in one minibatch"
    )
    parser.add_argument(
        "--learning_rate", type=float, default=2e-3, help="Learning rate to use."
    )
    parser.add_argument(
        "--model",
        default="super",
        choices=["super", "tanh", "circ", "logistic", "circ_dist"],
        help="Model to use for training.",
    )
    parser.add_argument(
        "--only_first_spike",
        type=bool,
        default=False,
        help="Only one spike per input (latency coding).",
    )
    parser.add_argument(
        "--optimizer",
        type=str,
        default="adam",
        choices=["adam", "sgd"],
        help="Optimizer to use for training.",
    )
    parser.add_argument(
        "--train_only_output",
        type=bool,
        default=False,
        help="Train only the last layer",
    )
    parser.add_argument(
        "--seq_length", default=200, type=int, help="Number of timesteps to do."
    )
    args = parser.parse_args()

    main(args)



import torch

from norse.torch.functional.lsnn import (
    LSNNState,
    LSNNFeedForwardState,
    lsnn_feed_forward_step,
    lsnn_step,
    ada_lif_step,
)


def test_lsnn_step():
    x = torch.ones(20)
    s = LSNNState(
        z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10), b=torch.zeros(10)
    )
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = lsnn_step(x, s, input_weights, recurrent_weights)


def test_lsnn_step_batch():
    x = torch.ones(16, 20)
    s = LSNNState(
        z=torch.zeros(16, 10),
        v=torch.zeros(16, 10),
        i=torch.zeros(16, 10),
        b=torch.zeros(16, 10),
    )
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = lsnn_step(x, s, input_weights, recurrent_weights)


def test_ada_lif_step():
    x = torch.ones(20)
    s = LSNNState(
        z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10), b=torch.zeros(10)
    )
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = ada_lif_step(x, s, input_weights, recurrent_weights)


def test_lsnn_feed_forward_step():
    x = torch.ones(10)
    s = LSNNFeedForwardState(v=torch.zeros(10), i=torch.zeros(10), b=torch.zeros(10))

    for _ in range(100):
        _, s = lsnn_feed_forward_step(x, s)



import torch
from norse.torch.functional.heaviside import heaviside


class SuperSpike(torch.autograd.Function):
    """SuperSpike surrogate gradient as described in Section 3.3.2 of

    F. Zenke, S. Ganguli, "SuperSpike: Supervised Learning in
                           Multilayer Spiking Neural Networks",
    Neural Computation 30, 1514–1541 (2018),
    doi:10.1162/neco_a_01086
    """

    @staticmethod
    @torch.jit.ignore
    def forward(ctx, input_tensor: torch.Tensor, alpha: float) -> torch.Tensor:
        ctx.save_for_backward(input_tensor)
        ctx.alpha = alpha
        return heaviside(input_tensor)

    @staticmethod
    @torch.jit.ignore
    def backward(ctx, grad_output):
        (inp,) = ctx.saved_tensors
        alpha = ctx.alpha
        grad_input = grad_output.clone()
        grad = grad_input / (alpha * torch.abs(inp) + 1.0).pow(
            2
        )  # section 3.3.2 (beta -> alpha)
        return grad, None


@torch.jit.ignore
def super_fn(x: torch.Tensor, alpha: float = 100.0) -> torch.Tensor:
    return SuperSpike.apply(x, alpha)



import torch

from norse.torch.module.lif import LIFFeedForwardLayer
from norse.torch.module.lift import Lift


def test_lift_conv():
    batch_size = 16
    seq_length = 20
    in_channels = 64
    out_channels = 32
    conv2d = Lift(torch.nn.Conv2d(in_channels, out_channels, 5, 1))
    data = torch.randn(seq_length, batch_size, in_channels, 20, 30)
    output = conv2d(data)

    assert output.shape == torch.Size([seq_length, batch_size, out_channels, 16, 26])


def test_lift_sequential():
    batch_size = 16
    seq_length = 20
    in_channels = 64
    out_channels = 32

    data = torch.randn(seq_length, batch_size, in_channels, 20, 30)
    module = torch.nn.Sequential(
        Lift(torch.nn.Conv2d(in_channels, out_channels, 5, 1)),
        LIFFeedForwardLayer(),
    )
    output, _ = module(data)

    assert output.shape == torch.Size([seq_length, batch_size, out_channels, 16, 26])


if __name__ == "__main__":
    test_lift_conv()
    test_lift_sequential()



import torch
from pytest import raises


from norse.torch.functional.threshold import (
    threshold,
    sign,
    heavi_erfc_fn,
    logistic_fn,
    circ_dist_fn,
)


def test_heavi_erfc_fn_forward():
    assert torch.equal(heavi_erfc_fn(torch.ones(100), 100.0), torch.ones(100))
    assert torch.equal(heavi_erfc_fn(-1.0 * torch.ones(100), 100.0), torch.zeros(100))


def test_heavi_erfc_fn_backward():
    x = torch.ones(10, requires_grad=True)
    out = heavi_erfc_fn(x, 0.1)
    out.backward(torch.ones(10))

    assert torch.sum(x.grad > 0) == 10

    x = torch.ones(10, requires_grad=True)
    out = heavi_erfc_fn(-0.001 * x, 0.1)
    out.backward(torch.ones(10))

    assert torch.sum(x.grad < 0) == 10


def test_logistic_fn_forward():
    x = torch.ones(10)
    out = logistic_fn(x, 0.1)
    assert torch.sum(torch.logical_or(out == 1, out == 0)) == 10


def test_logistic_fn_backward():
    x = torch.ones(10, requires_grad=True)
    out = logistic_fn(x, 0.1)
    out.backward(torch.ones(10))

    assert torch.sum(x.grad > 0) == 10

    x = torch.ones(10, requires_grad=True)
    out = logistic_fn(-0.001 * x, 0.1)
    out.backward(torch.ones(10))

    assert torch.sum(x.grad < 0) == 10


def test_circ_dist_fn_forward():
    x = torch.ones(10)
    out = circ_dist_fn(x, 0.1)
    assert torch.sum(torch.logical_or(out == 1, out == 0)) == 10


def test_circ_dist_fn_backward():
    x = torch.ones(10, requires_grad=True)
    out = circ_dist_fn(x, 0.1)
    out.backward(torch.ones(10))

    assert torch.sum(x.grad > 0) == 10

    x = torch.ones(10, requires_grad=True)
    out = circ_dist_fn(-0.001 * x, 0.1)
    out.backward(torch.ones(10))

    assert torch.sum(x.grad < 0) == 10


def test_threshold_throws():
    alpha = 10.0
    x = torch.ones(10)

    with raises(ValueError):
        _ = threshold(x, "noasd", alpha)


def test_threshold_backward():
    alpha = 10.0
    x = torch.ones(10)

    methods = ["super", "tanh", "tent", "circ", "heavi_erfc"]

    for method in methods:
        x = torch.ones(10, requires_grad=True)
        out = threshold(x, method, alpha)
        out.backward(torch.ones(10))

        x = torch.full((10,), 0.1, requires_grad=True)
        out = threshold(x, method, alpha)
        out.backward(torch.ones(10))

        x = torch.full((10,), -0.1, requires_grad=True)
        out = threshold(x, method, alpha)
        out.backward(torch.ones(10))


def test_threshold():
    alpha = 10.0

    methods = ["super", "heaviside", "tanh", "tent", "circ", "heavi_erfc"]

    for method in methods:
        x = torch.ones(10)
        out = threshold(x, method, alpha)
        assert torch.equal(out, torch.ones(10))

        x = torch.full((10,), 0.1)
        out = threshold(x, method, alpha)
        assert torch.equal(out, torch.ones(10))

        x = torch.full((10,), -0.1)
        out = threshold(x, method, alpha)
        assert torch.equal(out, torch.zeros(10))


def test_sign():
    alpha = 10.0

    methods = [
        "super",
        "heaviside",
        "tanh",
        "tent",
        "circ",
    ]

    for method in methods:
        x = torch.ones(10)
        out = sign(x, method, alpha)
        assert torch.equal(out, torch.ones(10))

        x = torch.full((10,), 0.1)
        out = sign(x, method, alpha)
        assert torch.equal(out, torch.ones(10))

        x = torch.full((10,), -0.1)
        out = sign(x, method, alpha)
        assert torch.equal(out, -torch.ones(10))



import torch

from norse.torch.module.lif_adex import (
    LIFAdExCell,
    LIFAdExLayer,
    LIFAdExFeedForwardCell,
    LIFAdExFeedForwardState,
)


def test_lif_adex_cell():
    cell = LIFAdExCell(2, 4)
    data = torch.randn(5, 2)
    out, _ = cell(data)

    assert out.shape == (5, 4)


def test_lif_adex_repr():
    cell = LIFAdExCell(2, 4)
    assert (
        str(cell)
        == "LIFAdExCell(2, 4, p=LIFAdExParameters(adaptation_current=tensor(4), adaptation_spike=tensor(0.0200), delta_T=tensor(0.5000), tau_ada_inv=tensor(2.), tau_syn_inv=tensor(200.), tau_mem_inv=tensor(100.), v_leak=tensor(0.), v_th=tensor(1.), v_reset=tensor(0.), method='super', alpha=100.0), dt=0.001)"
    )


def test_lif_adex_cell_backward():
    cell = LIFAdExCell(2, 4)
    data = torch.randn(5, 2)
    out, _ = cell(data)
    out.sum().backward()


def test_lif_adex_layer():
    layer = LIFAdExLayer(2, 4)
    data = torch.randn(10, 5, 2)
    out, _ = layer(data)

    assert out.shape == (10, 5, 4)


def test_lif_adex_feedforward_cell():
    layer = LIFAdExFeedForwardCell()
    data = torch.randn(5, 2, 4)
    out, _ = layer(data)

    assert out.shape == (5, 2, 4)


def test_lif_adex_feedforward_cell_state():
    cell = LIFAdExFeedForwardCell()
    input_tensor = torch.randn(5, 2, 4)

    state = LIFAdExFeedForwardState(
        v=cell.p.v_leak,
        i=torch.zeros(
            input_tensor.shape,
        ),
        a=torch.zeros(
            input_tensor.shape,
        ),
    )

    out, _ = cell(input_tensor, state)

    assert out.shape == (5, 2, 4)


def test_lif_adex_feedforward_repr():
    cell = LIFAdExFeedForwardCell()
    assert (
        str(cell)
        == "LIFAdExFeedForwardCell(p=LIFAdExParameters(adaptation_current=tensor(4), adaptation_spike=tensor(0.0200), delta_T=tensor(0.5000), tau_ada_inv=tensor(2.), tau_syn_inv=tensor(200.), tau_mem_inv=tensor(100.), v_leak=tensor(0.), v_th=tensor(1.), v_reset=tensor(0.), method='super', alpha=100.0), dt=0.001)"
    )


def test_lif_adex_feedforward_cell_backward():
    cell = LIFAdExFeedForwardCell()
    data = torch.randn(5, 2, 4)
    out, _ = cell(data)
    out.sum().backward()



import torch

from norse.torch.module.lif import (
    LIFCell,
    LIFLayer,
    LIFFeedForwardCell,
    LIFFeedForwardLayer,
)


class SNNetwork(torch.nn.Module):
    def __init__(self):
        super(SNNetwork, self).__init__()
        self.l0 = LIFCell(12, 6)
        self.l1 = LIFCell(6, 1)
        self.s0 = self.s1 = None

    def forward(self, spikes):
        spikes, self.s0 = self.l0(spikes, self.s0)
        _, self.s1 = self.l1(spikes, self.s1)
        return self.s1.v.squeeze()


def test_lif_cell():
    cell = LIFCell(2, 4)
    data = torch.randn(5, 2)
    out, s = cell(data)

    for x in s:
        assert x.shape == (5, 4)
    assert out.shape == (5, 4)


def test_lif_layer():
    layer = LIFLayer(2, 4)
    data = torch.randn(10, 5, 2)
    out, _ = layer(data)

    assert out.shape == (10, 5, 4)


def test_lif_cell_sequence():
    l1 = LIFCell(8, 6)
    l2 = LIFCell(6, 4)
    l3 = LIFCell(4, 1)
    z = torch.ones(10, 8)
    z, s1 = l1(z)
    z, s2 = l2(z)
    z, s3 = l3(z)
    assert s1.v.shape == (10, 6)
    assert s2.v.shape == (10, 4)
    assert s3.v.shape == (10, 1)
    assert z.shape == (10, 1)


def test_lif_cell_repr():
    cell = LIFCell(8, 6)
    assert (
        str(cell)
        == "LIFCell(8, 6, p=LIFParameters(tau_syn_inv=tensor(200.), tau_mem_inv=tensor(100.), v_leak=tensor(0.), v_th=tensor(1.), v_reset=tensor(0.), method='super', alpha=tensor(100.)), dt=0.001)"
    )


def test_lif_feedforward_cell():
    layer = LIFFeedForwardCell()
    data = torch.randn(5, 4)
    out, s = layer(data)

    assert out.shape == (5, 4)
    for x in s:
        assert x.shape == (5, 4)


def test_lif_feedforward_cell_backward():
    # Tests that gradient variables can be used in subsequent applications
    cell = LIFFeedForwardCell()
    data = torch.randn(5, 4)
    out, s = cell(data)
    out, _ = cell(out, s)
    loss = out.sum()
    loss.backward()


def test_lif_feedforward_layer():
    layer = LIFFeedForwardLayer()
    data = torch.randn(10, 5, 4)
    out, s = layer(data)
    assert out.shape == (10, 5, 4)
    for x in s:
        assert x.shape == (5, 4)


def test_backward():
    model = LIFCell(12, 1)
    data = torch.ones(100, 12)
    out, _ = model(data)
    loss = out.sum()
    loss.backward()


def test_backward_iteration():
    # Tests that gradient variables can be used in subsequent applications
    model = LIFCell(6, 6)
    data = torch.ones(100, 6)
    out, s = model(data)
    out, _ = model(out, s)
    loss = out.sum()
    loss.backward()


def test_backward_model():
    model = SNNetwork()
    data = torch.ones(100, 12)
    out = model(data)
    loss = out.sum()
    loss.backward()



from typing import List, NamedTuple

import numpy as np


class BenchmarkConfig(NamedTuple):
    """
    Benchmark configurations for the benchmarking setup
    """

    batch_size: int
    device: str
    dt: float
    label: str
    runs: int
    profile: bool
    sequence_length: int
    start: int
    stop: int
    step: int


class BenchmarkParameters(NamedTuple):
    """
    Benchmark parameters used as input for networks
    """

    device: str
    dt: float
    features: int
    batch_size: int
    sequence_length: int


class BenchmarkData(NamedTuple):
    """
    Result from a benchmark run N number of times
    """

    config: BenchmarkConfig
    durations: np.ndarray
    parameters: BenchmarkParameters



import torch

from norse.torch.functional.lif import LIFFeedForwardState, lif_feed_forward_step
from norse.torch.functional.regularization import regularize_step, voltage_accumulator


def test_regularisation_spikes():
    x = torch.ones(5, 10)
    s = LIFFeedForwardState(torch.ones(10), torch.ones(10))
    z, s = lif_feed_forward_step(x, s)
    zr, rs = regularize_step(z, s)
    assert torch.equal(z, zr)
    assert rs == 0
    z, s = lif_feed_forward_step(x, s)
    zr, rs = regularize_step(z, s)
    assert rs == 50


def test_regularisation_voltage():
    x = torch.ones(5, 10)
    s = LIFFeedForwardState(torch.ones(10), torch.ones(10))
    z, s = lif_feed_forward_step(x, s)
    # pytype: disable=wrong-arg-types
    zr, rs = regularize_step(z, s, accumulator=voltage_accumulator)
    # pytype: enable=wrong-arg-types
    assert torch.equal(z, zr)
    assert torch.equal(s.v, rs)


def test_regularisation_voltage_state():
    x = torch.ones(5, 10)
    state = torch.zeros(10)
    s = LIFFeedForwardState(torch.ones(10), torch.ones(10))
    z, s = lif_feed_forward_step(x, s)
    # pytype: disable=wrong-arg-types
    zr, rs = regularize_step(z, s, accumulator=voltage_accumulator, state=state)
    # pytype: enable=wrong-arg-types
    assert torch.equal(z, zr)
    assert torch.equal(s.v, rs)



import argparse
import matplotlib.pyplot as plt

# pytype: disable=import-error
import pandas as pd

# pytype: enable=import-error
import math


def plot_frames(frames, title):
    render_frames(frames, title)
    plt.show()


def save_frames(frames, title, filename):
    render_frames(frames, title)
    plt.savefig(filename)


def render_frames(frames, title):
    ax = plt.gca(yscale="log")
    for frame in frames:
        label = frame["label"][0].replace("_lif", "")
        plt.fill_between(
            frame["input_features"],
            frame["duration_mean"] - frame["duration_std"] * 2,
            frame["duration_mean"] + frame["duration_std"] * 2,
            alpha=0.2,
        )
        frame.plot(y="duration_mean", x="input_features", ax=ax, label=label)
        # Plot the crash, if any
        is_na = frame["duration_mean"].isnull()
        if is_na.any():
            last_index = is_na[is_na == True].index[0] - 1
            last_value = frame.loc[last_index]
            plt.scatter(
                x=[last_value["input_features"]],
                y=[last_value["duration_mean"]],
                color=ax.lines[-1].get_color(),
                s=60,
            )

    xmin = frame["input_features"].min()
    xmax = frame["input_features"].max()

    ax.set_xlim(math.floor(xmin / 1000), math.ceil(xmax / 1000) * 1000)
    ax.set_title(title)
    ax.set_xlabel("No. of features")
    ax.set_ylabel("Running time in seconds")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Plot a number of csv benchark files against each other"
    )
    parser.add_argument("files", type=argparse.FileType("r"), nargs="+")
    parser.add_argument(
        "--to",
        type=str,
        required=False,
        help="Save to given file instead of displaying",
    )
    parser.add_argument("--title", type=str, default="", help="Figure title")
    args = parser.parse_args()

    files = args.files
    dfs = []
    for f in files:
        dfs.append(pd.read_csv(f))
    if args.to:
        save_frames(dfs, args.title, args.to)
    else:
        plot_frames(dfs, args.title)



import pytest
import torch

from ...functional.stdp import (
    STDPParameters,
    STDPState,
    stdp_step_linear,
    stdp_step_conv2d,
)


def create_id(param):
    return param[0]


# Linear STDP
@pytest.fixture(
    scope="function",
    params=[
        ["additive", 0.0],
        ["additive_step", 0.0],
        ["multiplicative_pow", 0.75],
        ["multiplicative_relu", 1.0],
    ],
    ids=create_id,
)
def initialise_for_linear_stdp(request):
    stdp_algorithm = request.param[0]
    n_batches = 1
    n_pre, n_post = 2, 3
    w0 = 0.5 * torch.ones(n_post, n_pre)
    z_pre = torch.tensor(
        [
            [[1.0, 1.0]],
            [[0.0, 0.0]],
            [[1.0, 1.0]],
        ]
    ).float()
    z_post = torch.tensor(
        [
            [[0.0, 0.0, 0.0]],
            [[1.0, 1.0, 1.0]],
            [[0.0, 0.0, 0.0]],
        ]
    ).float()
    state_stdp = STDPState(
        t_pre=torch.zeros(n_batches, n_pre),
        t_post=torch.zeros(n_batches, n_post),
    )
    p_stdp = STDPParameters(
        eta_minus=1e-1,
        eta_plus=3e-1,  # Best to check with large, asymmetric learning-rates
        stdp_algorithm=stdp_algorithm,
        mu=request.param[1],
        hardbound=False,
        convolutional=False,
    )

    return (
        stdp_algorithm,
        n_batches,
        n_pre,
        n_post,
        w0,
        z_pre,
        z_post,
        state_stdp,
        p_stdp,
    )


def test_linear_stdp_stepper(initialise_for_linear_stdp):
    (
        _,
        n_batches,
        n_pre,
        n_post,
        w,
        z_pre,
        z_post,
        state_stdp,
        p_stdp,
    ) = initialise_for_linear_stdp

    n_time = z_pre.shape[0]

    t_pre = 0.0
    t_post = 0.0
    for n_t in range(n_time):
        w0 = w
        w, state_stdp = stdp_step_linear(
            z_pre[n_t],
            z_post[n_t],
            w,
            state_stdp,
            p_stdp,
            dt=0.001,
        )

        # Calculating the gradient for one synapse
        t_pre += (
            0.001 * (p_stdp.tau_pre_inv) * (-t_pre + p_stdp.a_pre * z_pre[n_t][0][0])
        )
        t_post += (
            0.001
            * (p_stdp.tau_post_inv)
            * (-t_post + p_stdp.a_post * z_post[n_t][0][0])
        )

        # Check potentiation
        if n_t == 1:
            assert torch.allclose(
                torch.abs(w - w0) / torch.pow(p_stdp.w_max - w0, p_stdp.mu),
                p_stdp.eta_plus * t_pre,
                atol=1e-6,
                rtol=0.0,
            )
        # Check depression
        if n_t == 2:
            assert torch.allclose(
                torch.abs(w - w0) / torch.pow(w0, p_stdp.mu),
                p_stdp.eta_minus * t_post,
                atol=1e-6,
                rtol=0.0,
            )

    # Shape checks
    assert state_stdp.t_pre.shape == (n_batches, n_pre)
    assert state_stdp.t_post.shape == (n_batches, n_post)
    assert w.shape == (n_post, n_pre)


# Conv2D STDP
@pytest.fixture(
    scope="function",
    params=[
        ["additive", 0.0],
        ["additive_step", 0.0],
        ["multiplicative_pow", 0.75],
        ["multiplicative_relu", 1.0],
    ],
    ids=create_id,
)
def initialise_for_conv2d_stdp(request):
    stdp_algorithm = request.param[0]
    n_batches = 1
    c_pre, c_post = 3, 2
    hw_pre, hw_post = (10, 10), (8, 8)
    hw_kern = (3, 3)
    w0 = torch.nn.Conv2d(c_pre, c_post, *hw_kern).weight.detach()
    torch.nn.init.constant_(w0, 0.5)
    z_pre = torch.stack(
        (
            torch.ones(n_batches, c_pre, *hw_pre),
            torch.zeros(n_batches, c_pre, *hw_pre),
            torch.ones(n_batches, c_pre, *hw_pre),
        ),
        dim=0,
    )
    z_post = torch.stack(
        (
            torch.zeros(n_batches, c_post, *hw_post),
            torch.ones(n_batches, c_post, *hw_post),
            torch.zeros(n_batches, c_post, *hw_post),
        ),
        dim=0,
    )
    state_stdp = STDPState(
        t_pre=torch.zeros(n_batches, c_pre, *hw_pre),
        t_post=torch.zeros(n_batches, c_post, *hw_post),
    )
    p_stdp = STDPParameters(
        eta_minus=1e-2,
        eta_plus=3e-2,  # Best to check with large, asymmetric learning-rates
        stdp_algorithm=stdp_algorithm,
        mu=request.param[1],
        hardbound=False,
        convolutional=True,
    )

    return (
        stdp_algorithm,
        n_batches,
        c_pre,
        c_post,
        hw_pre,
        hw_post,
        hw_kern,
        w0,
        z_pre,
        z_post,
        state_stdp,
        p_stdp,
    )


def test_conv2d_stdp_stepper(initialise_for_conv2d_stdp):
    (
        _,
        n_batches,
        c_pre,
        c_post,
        hw_pre,
        hw_post,
        hw_kern,
        w,
        z_pre,
        z_post,
        state_stdp,
        p_stdp,
    ) = initialise_for_conv2d_stdp

    n_time = z_pre.shape[0]

    t_pre = 0.0
    t_post = 0.0
    for n_t in range(n_time):
        w0 = w
        w, state_stdp = stdp_step_conv2d(
            z_pre[n_t],
            z_post[n_t],
            w,
            state_stdp,
            p_stdp,
            dt=0.001,
        )

        # Calculating the gradient for one synapse
        t_pre += (
            0.001
            * p_stdp.tau_pre_inv
            * (-t_pre + p_stdp.a_pre * z_pre[n_t][0][0][0][0])
        )
        t_post += (
            0.001
            * p_stdp.tau_post_inv
            * (-t_post + p_stdp.a_post * z_post[n_t][0][0][0][0])
        )

        # Check potentiation
        if n_t == 1:
            assert torch.allclose(
                torch.abs(w - w0) / torch.pow(p_stdp.w_max - w0, p_stdp.mu),
                hw_post[0] * hw_post[1] * p_stdp.eta_plus * t_pre,
                atol=1e-6,
                rtol=0.0,
            )

        # Check depression
        if n_t == 2:
            assert torch.allclose(
                torch.abs(w - w0) / torch.pow(w0 - p_stdp.w_min, p_stdp.mu),
                hw_post[0] * hw_post[1] * p_stdp.eta_minus * t_post,
                atol=1e-6,
                rtol=0.0,
            )

    # Shape checks
    assert state_stdp.t_pre.shape == (n_batches, c_pre, *hw_pre)
    assert state_stdp.t_post.shape == (n_batches, c_post, *hw_post)
    assert w.shape == (c_post, c_pre, *hw_kern)



from typing import Optional, Tuple

import numpy as np
import torch

from ..functional.lif_ex import (
    LIFExState,
    LIFExFeedForwardState,
    LIFExParameters,
    lif_ex_step,
    lif_ex_feed_forward_step,
)


class LIFExCell(torch.nn.Module):
    r"""Computes a single euler-integration step of a recurrent
    exponential LIF neuron-model adapted from
    https://neuronaldynamics.epfl.ch/online/Ch5.S2.html.
    More specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} \left(v_{\text{leak}} - v + i + \Delta_T exp\left({{v - v_{\text{th}}} \over {\Delta_T}}\right)\right) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\
            i &= i + w_{\text{input}} z_{\text{in}} \\
            i &= i + w_{\text{rec}} z_{\text{rec}}
        \end{align*}

    where :math:`z_{\text{rec}}` and :math:`z_{\text{in}}` are the
    recurrent and input spikes respectively.

    Parameters:
        input_size (int): Size of the input.
        hidden_size (int): Size of the hidden state.
        p (LIFExParameters): Parameters of the LIF neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif_ex = LIFExCell(10, 20)
        >>> input = torch.randn(batch_size, 10)
        >>> output, s0 = lif_ex(input)
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LIFExParameters = LIFExParameters(),
        dt: float = 0.001,
    ):
        super(LIFExCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) * np.sqrt(2 / hidden_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) * np.sqrt(2 / hidden_size)
        )
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def extra_repr(self):
        s = f"{self.input_size}, {self.hidden_size}, p={self.p}, dt={self.dt}"
        return s

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFExState] = None
    ) -> Tuple[torch.Tensor, LIFExState]:
        if state is None:
            state = LIFExState(
                z=torch.zeros(
                    (input_tensor.shape[0], self.hidden_size),
                    dtype=input_tensor.dtype,
                    device=input_tensor.device,
                ),
                v=self.p.v_leak.detach(),
                i=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.v.requires_grad = True
        return lif_ex_step(
            input_tensor,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )


class LIFExLayer(torch.nn.Module):
    """
    A neuron layer that wraps a recurrent LIFExCell in time such
    that the layer keeps track of temporal sequences of spikes.
    After application, the layer returns a tuple containing
      (spikes from all timesteps, state from the last timestep).

    Example:
        >>> data = torch.zeros(10, 5, 2) # 10 timesteps, 5 batches, 2 neurons
        >>> l = LIFExLayer(2, 4)
        >>> l(data) # Returns tuple of (Tensor(10, 5, 4), LIFExState)
    """

    def __init__(self, *cell_args, **kw_args):
        super(LIFExLayer, self).__init__()
        self.cell = LIFExCell(*cell_args, **kw_args)

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFExState] = None
    ) -> Tuple[torch.Tensor, LIFExState]:
        inputs = input_tensor.unbind(0)
        outputs = []  # torch.jit.annotate(List[torch.Tensor], [])
        for _, input_step in enumerate(inputs):
            out, state = self.cell(input_step, state)
            outputs += [out]
        # pytype: disable=bad-return-type
        return torch.stack(outputs), state
        # pytype: enable=bad-return-type


class LIFExFeedForwardCell(torch.nn.Module):
    r"""Computes a single euler-integration step of a feed-forward exponential
    LIF neuron-model adapted from
    https://neuronaldynamics.epfl.ch/online/Ch5.S2.html.
    It takes as input the input current as generated by an arbitrary torch
    module or function. More specifically it implements one integration step
    of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} \left(v_{\text{leak}} - v + i + \Delta_T exp\left({{v - v_{\text{th}}} \over {\Delta_T}}\right)\right) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        i = i + i_{\text{in}}

    where :math:`i_{\text{in}}` is meant to be the result of applying
    an arbitrary pytorch module (such as a convolution) to input spikes.

    Parameters:
        p (LIFExParameters): Parameters of the LIFEx neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif_ex = LIFExFeedForwardCell(
        >>> data = torch.randn(batch_size, 20, 30)
        >>> output, s0 = lif_ex(data)
    """

    def __init__(self, p: LIFExParameters = LIFExParameters(), dt: float = 0.001):
        super(LIFExFeedForwardCell, self).__init__()
        self.p = p
        self.dt = dt

    def extra_repr(self):
        s = f"p={self.p}, dt={self.dt}"
        return s

    def forward(
        self, x: torch.Tensor, state: Optional[LIFExFeedForwardState] = None
    ) -> Tuple[torch.Tensor, LIFExFeedForwardState]:
        if state is None:
            state = LIFExFeedForwardState(
                v=self.p.v_leak.detach(),
                i=torch.zeros(*x.shape, device=x.device, dtype=x.dtype),
            )
            state.v.requires_grad = True
        return lif_ex_feed_forward_step(x, state, p=self.p, dt=self.dt)



import torch
import pytest

import norse
import norse.torch.functional.lif as lif_module
from norse.torch.functional.lif import (
    LIFState,
    LIFFeedForwardState,
    LIFParameters,
    LIFParametersJIT,
    lif_step,
    lif_feed_forward_step,
    _lif_feed_forward_step_jit,
    lif_current_encoder,
)


@pytest.fixture(autouse=True)
def cpp_fixture():
    setattr(norse, "IS_OPS_LOADED", True)  # Enable cpp


@pytest.fixture()
def jit_fixture():
    setattr(norse, "IS_OPS_LOADED", False)  # Disable cpp


def test_lif_cpp_and_jit_step():
    x = torch.ones(20)
    s = LIFState(z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10))
    input_weights = torch.linspace(0, 0.5, 200).view(10, 20)
    recurrent_weights = torch.linspace(0, -2, 100).view(10, 10)

    results = [
        torch.as_tensor([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        torch.as_tensor([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        torch.as_tensor([0, 0, 0, 0, 1, 1, 1, 1, 1, 1]),
        torch.as_tensor([0, 0, 1, 1, 1, 1, 1, 1, 1, 1]),
        torch.as_tensor([0, 0, 0, 0, 0, 0, 1, 1, 1, 1]),
        torch.as_tensor([0, 0, 0, 0, 1, 1, 0, 0, 0, 0]),
        torch.as_tensor([0, 0, 0, 0, 0, 0, 0, 1, 1, 1]),
        torch.as_tensor([0, 0, 0, 1, 0, 1, 1, 0, 1, 1]),
        torch.as_tensor([0, 1, 1, 0, 1, 0, 0, 1, 1, 1]),
        torch.as_tensor([0, 0, 0, 0, 0, 1, 1, 0, 0, 1]),
    ]

    cpp_results = []
    cpp_states = []
    for result in results:
        z, s = lif_step(x, s, input_weights, recurrent_weights)
        cpp_results.append(z)
        cpp_states.append(s)

    s = LIFState(z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10))
    setattr(norse, "IS_OPS_LOADED", False)  # Disable cpp

    for i, result in enumerate(results):
        z, s = lif_step(x, s, input_weights, recurrent_weights)
        assert torch.equal(z, result.float())
        assert torch.equal(z, cpp_results[i])
        assert torch.equal(s.v, cpp_states[i].v)
        assert torch.equal(s.z, cpp_states[i].z)
        assert torch.equal(s.i, cpp_states[i].i)


def test_lif_cpp_back(cpp_fixture):
    x = torch.ones(2)
    s = LIFState(z=torch.zeros(1), v=torch.zeros(1), i=torch.zeros(1))
    s.v.requires_grad = True
    input_weights = torch.ones(2)
    recurrent_weights = torch.ones(1)
    _, s = lif_step(x, s, input_weights, recurrent_weights)
    z, s = lif_step(x, s, input_weights, recurrent_weights)
    z.sum().backward()


def test_lif_jit_back(jit_fixture):
    x = torch.ones(2)
    s = LIFState(z=torch.zeros(1), v=torch.zeros(1), i=torch.zeros(1))
    s.v.requires_grad = True
    input_weights = torch.ones(2)
    recurrent_weights = torch.ones(1)
    _, s = lif_step(x, s, input_weights, recurrent_weights)
    z, s = lif_step(x, s, input_weights, recurrent_weights)
    z.sum().backward()


def test_lif_heavi():
    x = torch.ones(2, 1)
    s = LIFState(z=torch.ones(2, 1), v=torch.zeros(2, 1), i=torch.zeros(2, 1))
    input_weights = torch.ones(1, 1) * 10
    recurrent_weights = torch.ones(1, 1)
    p = LIFParameters(method="heaviside")
    _, s = lif_step(x, s, input_weights, recurrent_weights, p)
    z, s = lif_step(x, s, input_weights, recurrent_weights, p)
    assert z.max() > 0
    assert z.shape == (2, 1)


def test_lif_feed_forward_step():
    x = torch.ones(10)
    s = LIFFeedForwardState(v=torch.zeros(10), i=torch.zeros(10))

    results = [0.0, 0.1, 0.27, 0.487, 0.7335, 0.9963, 0.0, 0.3951, 0.7717, 0.0]

    for result in results:
        _, s = lif_feed_forward_step(x, s)
        assert torch.allclose(torch.as_tensor(result), s.v, atol=1e-4)


def test_lif_feed_forward_step_batch():
    x = torch.ones(2, 1)
    s = LIFFeedForwardState(v=torch.zeros(2, 1), i=torch.zeros(2, 1))

    z, s = lif_feed_forward_step(x, s)
    assert z.shape == (2, 1)


def test_lif_feed_forward_step_jit():
    x = torch.ones(10)
    s = LIFFeedForwardState(v=torch.zeros(10), i=torch.zeros(10))

    p = LIFParametersJIT(
        torch.as_tensor(1.0 / 5e-3),
        torch.as_tensor(1.0 / 1e-2),
        torch.as_tensor(0.0),
        torch.as_tensor(1.0),
        torch.as_tensor(0.0),
        "super",
        torch.as_tensor(0.0),
    )

    results = [0.0, 0.1, 0.27, 0.487, 0.7335, 0.9963, 0.0, 0.3951, 0.7717, 0.0]

    for result in results:
        _, s = _lif_feed_forward_step_jit(x, s, p)
        assert torch.allclose(torch.as_tensor(result), s.v, atol=1e-4)


def test_lif_current_encoder():
    x = torch.ones(10)
    v = torch.zeros(10)

    results = [
        0.1,
        0.19,
        0.2710,
        0.3439,
        0.4095,
        0.4686,
        0.5217,
        0.5695,
        0.6126,
        0.6513,
    ]

    for result in results:
        _, v = lif_current_encoder(x, v)
        assert torch.allclose(torch.as_tensor(result), v, atol=1e-4)



from typing import Optional, Tuple
import numpy as np
import torch

from ..functional.lif import LIFState, LIFFeedForwardState

from ..functional.lif_refrac import (
    LIFRefracParameters,
    LIFRefracState,
    LIFRefracFeedForwardState,
    lif_refrac_step,
    lif_refrac_feed_forward_step,
)


class LIFRefracCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a LIF
    neuron-model with absolute refractory period. More specifically it
    implements one integration step of the following ODE.

    .. math::
        \\begin{align*}
            \\dot{v} &= 1/\\tau_{\\text{mem}} (1-\\Theta(\\rho)) \
            (v_{\\text{leak}} - v + i) \\\\
            \\dot{i} &= -1/\\tau_{\\text{syn}} i \\\\
            \\dot{\\rho} &= -1/\\tau_{\\text{refrac}} \\Theta(\\rho)
        \\end{align*}

    together with the jump condition

    .. math::
        \\begin{align*}
            z &= \\Theta(v - v_{\\text{th}}) \\\\
            z_r &= \\Theta(-\\rho)
        \\end{align*}

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            i &= i + w_{\\text{input}} z_{\\text{in}} \\\\
            i &= i + w_{\\text{rec}} z_{\\text{rec}} \\\\
            \\rho &= \\rho + z_r \\rho_{\\text{reset}}
        \\end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the
    recurrent and input spikes respectively.

    Parameters:
        input_size (int): Size of the input. Also known as the number of input features.
        hidden_size (int): Size of the hidden state. Also known as the number of input features.
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use

    Examples:

        >>> batch_size = 16
        >>> lif = LIFRefracCell(10, 20)
        >>> input = torch.randn(batch_size, 10)
        >>> output, s0 = lif(input)
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LIFRefracParameters = LIFRefracParameters(),
        dt: float = 0.001,
    ):
        super(LIFRefracCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFRefracState] = None
    ) -> Tuple[torch.Tensor, LIFRefracState]:
        if state is None:
            state = LIFRefracState(
                LIFState(
                    z=torch.zeros(
                        input_tensor.shape[0],
                        self.hidden_size,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                    v=self.p.lif.v_leak.detach(),
                    i=torch.zeros(
                        input_tensor.shape[0],
                        self.hidden_size,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                ),
                rho=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.lif.v.requires_grad = True
        return lif_refrac_step(
            input_tensor,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )


class LIFRefracFeedForwardCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a
    LIF neuron-model with absolute refractory period. More specifically
    it implements one integration step of the following ODE.

    .. math::
        \\begin{align*}
            \\dot{v} &= 1/\\tau_{\\text{mem}} (1-\\Theta(\\rho)) \
            (v_{\\text{leak}} - v + i) \\\\
            \\dot{i} &= -1/\\tau_{\\text{syn}} i \\\\
            \\dot{\\rho} &= -1/\\tau_{\\text{refrac}} \\Theta(\\rho)
        \\end{align*}

    together with the jump condition

    .. math::
        \\begin{align*}
            z &= \\Theta(v - v_{\\text{th}}) \\\\
            z_r &= \\Theta(-\\rho)
        \\end{align*}

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            \\rho &= \\rho + z_r \\rho_{\\text{reset}}
        \\end{align*}

    Parameters:
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use

    Examples:
        >>> batch_size = 16
        >>> lif = LIFRefracFeedForwardCell()
        >>> input = torch.randn(batch_size, 20, 30)
        >>> output, s0 = lif(input)
    """

    def __init__(
        self,
        p: LIFRefracParameters = LIFRefracParameters(),
        dt: float = 0.001,
    ):
        super(LIFRefracFeedForwardCell, self).__init__()
        self.p = p
        self.dt = dt

    def forward(
        self,
        input_tensor: torch.Tensor,
        state: Optional[LIFRefracFeedForwardState] = None,
    ) -> Tuple[torch.Tensor, LIFRefracFeedForwardState]:
        if state is None:
            state = LIFRefracFeedForwardState(
                LIFFeedForwardState(
                    v=self.p.lif.v_leak.detach(),
                    i=torch.zeros(
                        input_tensor.shape,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                ),
                rho=torch.zeros(
                    input_tensor.shape,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.lif.v.requires_grad = True
        return lif_refrac_feed_forward_step(input_tensor, state, p=self.p, dt=self.dt)



import time
import torch

from norse.torch.functional.lif import (
    LIFFeedForwardState,
    LIFParametersJIT,
    _lif_feed_forward_step_jit,
)
from norse.torch.module.encode import PoissonEncoder

from .benchmark import BenchmarkParameters


class LIFBenchmark(torch.jit.ScriptModule):
    def __init__(self, parameters):
        super(LIFBenchmark, self).__init__()
        self.fc = torch.nn.Linear(parameters.features, parameters.features, bias=False)
        self.dt = parameters.dt

    @torch.jit.script_method
    def forward(
        self, input_spikes: torch.Tensor, p: LIFParametersJIT, s: LIFFeedForwardState
    ):
        sequence_length, batch_size, features = input_spikes.shape
        # spikes = torch.jit.annotate(List[Tensor], [])
        spikes = torch.empty(
            (sequence_length, batch_size, features), device=input_spikes.device
        )

        for ts in range(sequence_length):
            x = self.fc(input_spikes[ts])
            z, s = _lif_feed_forward_step_jit(input_tensor=x, state=s, p=p, dt=self.dt)
            spikes[ts] = z

        return spikes


def lif_feed_forward_benchmark(parameters: BenchmarkParameters):
    with torch.no_grad():
        model = LIFBenchmark(parameters).to(parameters.device)
        input_spikes = PoissonEncoder(parameters.sequence_length, dt=parameters.dt)(
            0.3
            * torch.ones(
                parameters.batch_size, parameters.features, device=parameters.device
            )
        ).contiguous()
        p = LIFParametersJIT(
            tau_syn_inv=torch.as_tensor(1.0 / 5e-3),
            tau_mem_inv=torch.as_tensor(1.0 / 1e-2),
            v_leak=torch.as_tensor(0.0),
            v_th=torch.as_tensor(1.0),
            v_reset=torch.as_tensor(0.0),
            method="super",
            alpha=torch.as_tensor(0.0),
        )
        s = LIFFeedForwardState(
            v=p.v_leak,
            i=torch.zeros(
                parameters.batch_size,
                parameters.features,
                device=parameters.device,
            ),
        )
        start = time.time()
        model(input_spikes, p, s)
        end = time.time()
        duration = end - start
        return duration



import torch


@torch.jit.script
def heaviside(data):
    """
    A `heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_
    that truncates numbers <= 0 to 0 and everything else to 1.

    .. math::
        H[n]=\\begin{cases} 0, & n <= 0, \\ 1, & n \\g 0, \\end{cases}
    """
    return torch.gt(data, torch.as_tensor(0.0)).to(data.dtype)



import torch
from norse.torch.functional.heaviside import heaviside


def test_heaviside():
    assert torch.equal(heaviside(torch.ones(100)), torch.ones(100))
    assert torch.equal(heaviside(-1.0 * torch.ones(100)), torch.zeros(100))



"""Norse is a library for doing deep learning with spiking neural networks.
"""

from . import benchmark, dataset, task

# Attempt import of optimized code
try:
    import torch
    from pathlib import Path

    torch.ops.load_library(Path(__file__).resolve().parent.parent / "norse_op.so")
    import sys

    setattr(sys.modules[__name__], "IS_OPS_LOADED", True)
except OSError:
    import sys

    setattr(sys.modules[__name__], "IS_OPS_LOADED", False)

from .torch import functional, models, module

__all__ = ["task", "benchmark", "dataset", "functional", "models", "module"]



import torch

from norse.torch.functional.lif import LIFState, LIFFeedForwardState
from norse.torch.functional.lif_mc import lif_mc_step, lif_mc_feed_forward_step


def test_lif_mc_step():
    x = torch.ones(20)
    s = LIFState(z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10))
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()
    g_coupling = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = lif_mc_step(x, s, input_weights, recurrent_weights, g_coupling)


def test_lif_mc_feed_forward_step():
    x = torch.ones(10)
    s = LIFFeedForwardState(v=torch.zeros(10), i=torch.zeros(10))
    g_coupling = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = lif_mc_feed_forward_step(x, s, g_coupling)



import torch

from norse.torch.functional.coba_lif import (
    CobaLIFState,
    CobaLIFFeedForwardState,
    coba_lif_feed_forward_step,
    coba_lif_step,
)


def test_coba_lif_step():
    x = torch.ones(20)
    s = CobaLIFState(
        z=torch.zeros(10), v=torch.zeros(10), g_e=torch.zeros(10), g_i=torch.zeros(10)
    )
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = coba_lif_step(x, s, input_weights, recurrent_weights)


def test_coba_lif_feed_forward_step():
    x = torch.ones(10)
    s = CobaLIFFeedForwardState(
        v=torch.zeros(10), g_e=torch.zeros(10), g_i=torch.zeros(10)
    )

    for _ in range(100):
        _, s = coba_lif_feed_forward_step(x, s)



"""
Stateless decoding functionality for Norse, where events in time are converted to numerical representations.
Many of the decoders are simply mirrors of the `PyTorch Tensor aggregation functions <https://pytorch.org/docs/stable/tensors.html>`_,
with the difference that they typically aggregate the first (temporal) dimension. They are therefore useful to inject in :met:`lift`_ed 
modules or situations where you need pickled functions.
"""
import torch


def sum_decode(tensor: torch.Tensor, dimension: int = 0):
    """
    Sums the input tensor in the first dimension by default.
    """
    return tensor.sum(dimension)



from functools import partial
from typing import Callable

from absl import app
from absl import flags
from absl import logging

import numpy as np

# pytype: disable=import-error
import pandas as pd

# pytype: enable=import-error
import time
import gc
import torch

from .benchmark import *

FLAGS = flags.FLAGS

flags.DEFINE_integer(
    "batch_size",
    32,
    "Number of data points per batch",
)
flags.DEFINE_integer("start", 250, "Start of the number of inputs to sweep")
flags.DEFINE_integer("step", 250, "Steps in which to sweep over the number of inputs")
flags.DEFINE_integer("stop", 5000, "Number of inputs to sweep to")
flags.DEFINE_integer("sequence_length", 1000, "Number of timesteps to simulate")
flags.DEFINE_float("dt", 0.001, "Simulation timestep")
flags.DEFINE_enum("device", "cuda", ["cuda", "cpu"], "Device to use [cpu, cuda]")
flags.DEFINE_integer("runs", 20, "Number of runs per simulation step")
flags.DEFINE_bool("profile", False, "Profile Norse benchmark? (Only works for Norse)")

flags.DEFINE_bool("bindsnet", True, "Benchmark Bindsnet?")
flags.DEFINE_bool("genn", True, "Benchmark GeNN?")
flags.DEFINE_bool("norse", True, "Benchmark Norse?")


def benchmark(
    model: Callable[[BenchmarkParameters], float],
    collector: Callable[[BenchmarkData], dict],
    config: BenchmarkConfig,
):
    """
    Benchmarks a model with the given configurations
    """
    results = []
    for features in range(config.start, config.stop, config.step):
        parameters = BenchmarkParameters(
            device=config.device,
            dt=config.dt,
            features=features,
            batch_size=config.batch_size,
            sequence_length=config.sequence_length,
        )

        durations = []
        try:
            for _ in range(config.runs):
                duration = model(parameters)
                durations.append(duration)
                # Clean up by GC and empty cache
                # Thanks to https://github.com/BindsNET/bindsnet/issues/425#issuecomment-721780231
                gc.collect()
                with torch.no_grad():
                    torch.cuda.empty_cache()
        except KeyboardInterrupt:
            raise KeyboardInterrupt()
        except RuntimeError as e:
            message = f"RuntimeError when running benchmark {config} {parameters}: {e}"
            logging.error(message)

        data = BenchmarkData(
            config=config,
            durations=np.array(durations),
            parameters=parameters,
        )
        result = collector(data)

        logging.info(result)
        results += [result]

    return results


def collect(data: BenchmarkData, label: str) -> dict:
    return {
        "label": label,
        "input_features": data.parameters.features,
        "output_features": data.parameters.features,
        "batch_size": data.parameters.batch_size,
        "run": len(data.durations),
        "duration_mean": data.durations.mean(),
        "duration_std": data.durations.std(),
        "dt": data.parameters.dt,
        "time_steps": data.parameters.sequence_length,
        "device": data.config.device,
    }


def main(argv):
    # pytype: disable=import-error
    if FLAGS.bindsnet:
        import bindsnet_lif

        run_benchmark(bindsnet_lif.lif_feed_forward_benchmark, "BindsNET_lif")
    if FLAGS.genn:
        import genn_lif

        run_benchmark(genn_lif.lif_feed_forward_benchmark, "GeNN_lif")
    if FLAGS.norse:
        import norse_lif

        if FLAGS.profile:
            import torch.autograd.profiler as profiler

            with profiler.profile(
                profile_memory=True, use_cuda=(FLAGS.device == "cuda")
            ) as prof:
                run_benchmark(norse_lif.lif_feed_forward_benchmark, "Norse_lif")
            prof.export_chrome_trace("trace.json")
        else:
            run_benchmark(norse_lif.lif_feed_forward_benchmark, "Norse_lif")
    # pytype: enable=import-error


def run_benchmark(function, label):
    config = BenchmarkConfig(
        batch_size=FLAGS.batch_size,
        device=FLAGS.device,
        dt=FLAGS.dt,
        label=label,
        runs=FLAGS.runs,
        sequence_length=FLAGS.sequence_length,
        start=FLAGS.start,
        stop=FLAGS.stop,
        step=FLAGS.step,
        profile=FLAGS.profile,
    )

    collector = partial(collect, label=label)
    results = benchmark(function, collector, config)

    timestamp = time.strftime("%Y-%m-%d-%H-%M-%S")
    filename = f"{timestamp}-{label}.csv"
    pd.DataFrame(results).to_csv(filename)


if __name__ == "__main__":
    app.run(main)



import torch

from norse.torch.module.coba_lif import CobaLIFCell, CobaLIFState


def test_coba():
    cell = CobaLIFCell(4, 3)
    data = torch.ones(5, 4)
    spikes, state = cell(data)
    assert spikes.shape == (5, 3)
    assert state.v.shape == (5, 3)


def test_coba_state():
    cell = CobaLIFCell(4, 3)
    input_tensor = torch.ones(5, 4)
    state = CobaLIFState(
        z=torch.zeros(
            input_tensor.shape[0],
            cell.hidden_size,
            device=input_tensor.device,
            dtype=input_tensor.dtype,
        ),
        v=torch.zeros(
            input_tensor.shape[0],
            cell.hidden_size,
            device=input_tensor.device,
            dtype=input_tensor.dtype,
        ),
        g_e=torch.zeros(
            input_tensor.shape[0],
            cell.hidden_size,
            device=input_tensor.device,
            dtype=input_tensor.dtype,
        ),
        g_i=torch.zeros(
            input_tensor.shape[0],
            cell.hidden_size,
            device=input_tensor.device,
            dtype=input_tensor.dtype,
        ),
    )

    spikes, state = cell(input_tensor, state)
    assert spikes.shape == (5, 3)
    assert state.v.shape == (5, 3)


def test_coba_backward():
    cell = CobaLIFCell(4, 3)
    data = torch.ones(5, 4)
    spikes, _ = cell(data)
    spikes.sum().backward()



import torch

from norse.torch.functional.logical import (
    logical_and,
    logical_or,
    logical_xor,
    muller_c,
)
from norse.torch.functional.logical import posedge_detector


def test_logical_and():
    z = logical_and(torch.as_tensor([1, 0, 0, 1]), torch.as_tensor([1, 0, 0, 0]))
    assert torch.equal(z, torch.as_tensor([1, 0, 0, 0]))


def test_logical_or():
    z = logical_or(torch.as_tensor([1, 0, 0, 1]), torch.as_tensor([1, 0, 0, 0]))
    assert torch.equal(z, torch.as_tensor([1, 0, 0, 1]))


def test_logical_xor():
    z = logical_xor(torch.as_tensor([1, 0, 0, 1]), torch.as_tensor([1, 0, 0, 0]))
    assert torch.equal(z, torch.as_tensor([0, 0, 0, 1]))


def test_muller_c():
    z = muller_c(
        torch.as_tensor([1, 0, 0, 1]),
        torch.as_tensor([1, 0, 0, 0]),
        torch.as_tensor([1, 0, 0, 0]),
    )
    assert torch.equal(z, torch.as_tensor([1, 0, 0, 0]))


def test_posedge_detector():
    z = torch.as_tensor([1, 0, 0, 1])
    z_prev = torch.as_tensor([0, 0, 1, 1])
    z_out = posedge_detector(z, z_prev)
    assert torch.equal(z_out, torch.as_tensor([1, 0, 0, 0]))



import torch
import torch.jit

from typing import NamedTuple

from .heaviside import heaviside


@torch.jit.script
def pre_mask(weights, z):
    """Computes the mask produced by the pre-synaptic spikes on
    the synapse array."""
    return torch.transpose(
        torch.transpose(torch.zeros_like(weights), 1, 2) + z, 1, 2
    )  # pragma: no cover


@torch.jit.script
def post_mask(weights, z):
    """Computes the mask produced by post-synaptic spikes on
    the synapse array.
    """
    return torch.zeros_like(weights) + z  # pragma: no cover


@torch.jit.script
def post_pre_update(post_pre, post_spike_mask, pre_spike_mask):
    """Computes which synapses in the synapse array should be updated."""
    return heaviside(post_pre + post_spike_mask - pre_spike_mask)  # pragma: no cover


class CorrelationSensorParameters(NamedTuple):
    eta_p: torch.Tensor = torch.as_tensor(1.0)
    eta_m: torch.Tensor = torch.as_tensor(1.0)
    tau_ac_inv: torch.Tensor = torch.as_tensor(1.0 / 100e-3)
    tau_c_inv: torch.Tensor = torch.as_tensor(1.0 / 100e-3)


class CorrelationSensorState(NamedTuple):
    post_pre: torch.Tensor
    correlation_trace: torch.Tensor
    anti_correlation_trace: torch.Tensor


def correlation_sensor_step(
    z_pre: torch.Tensor,
    z_post: torch.Tensor,
    state: CorrelationSensorState,
    p: CorrelationSensorParameters = CorrelationSensorParameters(),
    dt: float = 0.001,
) -> CorrelationSensorState:
    """Euler integration step of an idealized version of the correlation sensor
    as it is present on the BrainScaleS 2 chips.
    """
    dcorrelation_trace = dt * p.tau_c_inv * (-state.correlation_trace)
    correlation_trace_decayed = (
        state.correlation_trace + (1 - state.post_pre) * dcorrelation_trace
    )

    danti_correlation_trace = dt * p.tau_ac_inv * (-state.anti_correlation_trace)
    anti_correlation_trace_decayed = (
        state.anti_correlation_trace + state.post_pre * danti_correlation_trace
    )

    # compute the pre and post masks based on the current spikes
    pre_spike_mask = pre_mask(state.post_pre, z_pre)
    post_spike_mask = post_mask(state.post_pre, z_post)

    post_pre_new = post_pre_update(state.post_pre, post_spike_mask, pre_spike_mask)
    correlation_trace_new = correlation_trace_decayed + (p.eta_p * pre_spike_mask)
    anti_correlation_trace_new = (
        anti_correlation_trace_decayed + p.eta_m * post_spike_mask
    )

    return CorrelationSensorState(
        post_pre=post_pre_new,
        correlation_trace=correlation_trace_new,
        anti_correlation_trace=anti_correlation_trace_new,
    )


def correlation_based_update(
    ts: int,
    linear_update: torch.nn.Module,
    weights: torch.Tensor,
    correlation_state: CorrelationSensorState,
    learning_rate: float,
    ts_frequency: int,
):
    if ts % ts_frequency == 0:
        (
            _,
            input_features,
            hidden_features,
        ) = correlation_state.correlation_trace.shape
        # proposed weight update
        dw = torch.cat(
            (
                correlation_state.correlation_trace.flatten(),
                correlation_state.anti_correlation_trace.flatten(),
            )
        )
        dw = linear_update(dw).detach()
        weights = weights + learning_rate * torch.reshape(
            dw, (hidden_features, input_features)
        )
        # reset correlation traces
        correlation_state.correlation_trace.zero_()
        correlation_state.anti_correlation_trace.zero_()

    return weights



"""
Modules for spiking neural network, adhering to the ``torch.nn.Module`` interface.
"""

from .coba_lif import CobaLIFCell, CobaLIFParameters, CobaLIFState
from .encode import (
    ConstantCurrentLIFEncoder,
    PoissonEncoder,
    PopulationEncoder,
    SignedPoissonEncoder,
    SpikeLatencyEncoder,
    SpikeLatencyLIFEncoder,
)
from .leaky_integrator import LICell, LIFeedForwardCell, LIParameters, LIState
from .lif import (
    LIFCell,
    LIFFeedForwardCell,
    LIFFeedForwardState,
    LIFParameters,
    LIFState,
    LIFLayer,
)
from .lif_adex import (
    LIFAdExCell,
    LIFAdExFeedForwardCell,
    LIFAdExFeedForwardState,
    LIFAdExParameters,
    LIFAdExState,
    LIFAdExLayer,
)
from .lif_correlation import (
    LIFCorrelation,
    LIFCorrelationParameters,
    LIFCorrelationState,
)
from .lif_ex import (
    LIFExCell,
    LIFExFeedForwardCell,
    LIFExFeedForwardState,
    LIFExLayer,
    LIFExParameters,
    LIFExState,
)
from .lif_mc import LIFMCCell
from .lif_mc_refrac import LIFMCRefracCell
from .lif_refrac import (
    LIFRefracCell,
    LIFRefracFeedForwardCell,
    LIFRefracFeedForwardState,
    LIFRefracParameters,
    LIFRefracState,
)
from .lift import Lift
from .lsnn import (
    LSNNCell,
    LSNNFeedForwardCell,
    LSNNFeedForwardState,
    LSNNParameters,
    LSNNState,
    LSNNLayer,
)
from .regularization import RegularizationCell
from .sequential import SequentialState

__all__ = [
    "CobaLIFCell",
    "CobaLIFState",
    "CobaLIFParameters",
    "ConstantCurrentLIFEncoder",
    "PoissonEncoder",
    "PopulationEncoder",
    "SignedPoissonEncoder",
    "SpikeLatencyEncoder",
    "SpikeLatencyLIFEncoder",
    "LICell",
    "LIFeedForwardCell",
    "LIParameters",
    "LIState",
    "LIFCell",
    "LIFLayer",
    "LIFFeedForwardCell",
    "LIFFeedForwardState",
    "LIFParameters",
    "LIFState",
    "LIFAdExCell",
    "LIFAdExFeedForwardCell",
    "LIFAdExFeedForwardState",
    "LIFAdExParameters",
    "LIFAdExState",
    "LIFAdExLayer",
    "LIFAdExCell",
    "LIFAdExFeedForwardCell",
    "LIFAdExFeedForwardState",
    "LIFAdExParameters",
    "LIFAdExState",
    "LIFCorrelation",
    "LIFCorrelationParameters",
    "LIFCorrelationState",
    "LIFExCell",
    "LIFExFeedForwardCell",
    "LIFExFeedForwardState",
    "LIFExLayer",
    "LIFExParameters",
    "LIFExState",
    "LIFMCCell",
    "LIFMCRefracCell",
    "LIFRefracParameters",
    "LIFRefracState",
    "LIFMCRefracCell",
    "LIFRefracParameters",
    "LIFRefracState",
    "LIFRefracCell",
    "LIFRefracFeedForwardCell",
    "LIFRefracFeedForwardState",
    "LIFRefracParameters",
    "LIFRefracState",
    "Lift",
    "LSNNCell",
    "LSNNFeedForwardCell",
    "LSNNFeedForwardState",
    "LSNNParameters",
    "LSNNLayer",
    "LSNNState",
    "RegularizationCell",
    "SequentialState",
]



import torch

from norse.torch.functional.lif_ex import (
    LIFExState,
    LIFExFeedForwardState,
    lif_ex_step,
    lif_ex_feed_forward_step,
    lif_ex_current_encoder,
)


def test_lif_ex_step():
    x = torch.ones(20)
    s = LIFExState(z=torch.zeros(10), v=torch.zeros(10), i=torch.zeros(10))
    input_weights = torch.randn(10, 20).float()
    recurrent_weights = torch.randn(10, 10).float()

    for _ in range(100):
        _, s = lif_ex_step(x, s, input_weights, recurrent_weights)


def test_lif_ex_feed_forward_step():
    x = torch.ones(10)
    s = LIFExFeedForwardState(v=torch.zeros(10), i=torch.zeros(10))

    for _ in range(100):
        _, s = lif_ex_feed_forward_step(x, s)


def test_lif_ex_current_encoder():
    x = torch.ones(10)
    v = torch.zeros(10)

    for _ in range(100):
        _, v = lif_ex_current_encoder(x, v)



from typing import NamedTuple, Tuple, overload

import torch

from norse.torch.functional.lif import (
    LIFParameters,
    LIFState,
    LIFFeedForwardState,
    lif_step,
    lif_feed_forward_step,
)
from norse.torch.functional.threshold import threshold


class LIFRefracState(NamedTuple):
    """State of a LIF neuron with absolute refractory period.

    Parameters:
        lif (LIFState): state of the LIF neuron integration
        rho (torch.Tensor): refractory state (count towards zero)
    """

    lif: LIFState
    rho: torch.Tensor


class LIFRefracParameters(NamedTuple):
    """Parameters of a LIF neuron with absolute refractory period.

    Parameters:
        lif (LIFParameters): parameters of the LIF neuron integration
        rho (torch.Tensor): refractory state (count towards zero)
    """

    lif: LIFParameters = LIFParameters()
    rho_reset: torch.Tensor = torch.as_tensor(5.0)


class LIFRefracFeedForwardState(NamedTuple):
    """State of a feed forward LIF neuron with absolute refractory period.

    Parameters:
        lif (LIFFeedForwardState): state of the feed forward LIF
                                   neuron integration
        rho (torch.Tensor): refractory state (count towards zero)
    """

    lif: LIFFeedForwardState
    rho: torch.Tensor


@overload
def compute_refractory_update(
    state: LIFRefracState,
    z_new: torch.Tensor,
    v_new: torch.Tensor,
    p: LIFRefracParameters = LIFRefracParameters(),
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    ...


@overload
def compute_refractory_update(
    state: LIFRefracFeedForwardState,
    z_new: torch.Tensor,
    v_new: torch.Tensor,
    p: LIFRefracParameters = LIFRefracParameters(),
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    ...


def compute_refractory_update(
    state,
    z_new: torch.Tensor,
    v_new: torch.Tensor,
    p: LIFRefracParameters = LIFRefracParameters(),
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Compute the refractory update.

    Parameters:
        state (LIFRefracState): Initial state of the refractory neuron.
        z_new (torch.Tensor): New spikes that were generated.
        v_new (torch.Tensor): New voltage after the lif update step.
        p (torch.Tensor): Refractoryp.
    """
    refrac_mask = threshold(state.rho, p.lif.method, p.lif.alpha)
    v_new = (1 - refrac_mask) * v_new + refrac_mask * state.lif.v
    z_new = (1 - refrac_mask) * z_new

    # compute update to refractory counter
    rho_new = (1 - z_new) * torch.nn.functional.relu(
        state.rho - refrac_mask
    ) + z_new * p.rho_reset

    return v_new, z_new, rho_new


def lif_refrac_step(
    input_tensor: torch.Tensor,
    state: LIFRefracState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LIFRefracParameters = LIFRefracParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFRefracState]:
    r"""Computes a single euler-integration step of a recurrently connected
     LIF neuron-model with a refractory period.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (LIFRefracState): state at the current time step
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use
    """
    z_new, s_new = lif_step(
        input_tensor, state.lif, input_weights, recurrent_weights, p.lif, dt
    )
    v_new, z_new, rho_new = compute_refractory_update(state, z_new, s_new.v, p)

    return z_new, LIFRefracState(LIFState(z_new, v_new, s_new.i), rho_new)


def lif_refrac_feed_forward_step(
    input_tensor: torch.Tensor,
    state: LIFRefracFeedForwardState,
    p: LIFRefracParameters = LIFRefracParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFRefracFeedForwardState]:
    r"""Computes a single euler-integration step of a feed forward
     LIF neuron-model with a refractory period.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (LIFRefracFeedForwardState): state at the current time step
        p (LIFRefracParameters): parameters of the lif neuron
        dt (float): Integration timestep to use
    """
    z_new, s_new = lif_feed_forward_step(input_tensor, state.lif, p.lif, dt)
    v_new, z_new, rho_new = compute_refractory_update(state, z_new, s_new.v, p)

    return (
        z_new,
        LIFRefracFeedForwardState(LIFFeedForwardState(v_new, s_new.i), rho_new),
    )



r"""
In this task, we train a spiking convolutional network to learn the
MNIST digit recognition task.
"""
import os
import uuid

from absl import app
from absl import flags
from absl import logging

import numpy as np
import matplotlib.pyplot as plt

import torch
import torch.utils.data
import torchvision

from norse.torch.models.conv import ConvNet4
from norse.torch.module.encode import ConstantCurrentLIFEncoder

FLAGS = flags.FLAGS

flags.DEFINE_bool(
    "only_first_spike", False, "Only one spike per input (latency coding)."
)
flags.DEFINE_bool("save_grads", False, "Save gradients of backward pass.")
flags.DEFINE_integer(
    "grad_save_interval", 10, "Interval for gradient saving of backward pass."
)
flags.DEFINE_bool("refrac", False, "Use refractory time.")
flags.DEFINE_integer("plot_interval", 10, "Interval for plotting.")
flags.DEFINE_float("input_scale", 1, "Scaling factor for input current.")
flags.DEFINE_bool(
    "find_learning_rate", False, "Use learning rate finder to find learning rate."
)
flags.DEFINE_enum("device", "cpu", ["cpu", "cuda"], "Device to use by pytorch.")
flags.DEFINE_integer("epochs", 10, "Number of training episodes to do.")
flags.DEFINE_integer("seq_length", 200, "Number of timesteps to do.")
flags.DEFINE_integer("batch_size", 32, "Number of examples in one minibatch.")
flags.DEFINE_enum(
    "model",
    "super",
    ["super", "tanh", "circ", "logistic", "circ_dist"],
    "Model to use for training.",
)
flags.DEFINE_string("prefix", "", "Prefix to use for saving the results")
flags.DEFINE_enum(
    "optimizer", "adam", ["adam", "sgd"], "Optimizer to use for training."
)
flags.DEFINE_bool("clip_grad", False, "Clip gradient during backpropagation")
flags.DEFINE_float("grad_clip_value", 1.0, "Gradient to clip at.")
flags.DEFINE_float("learning_rate", 2e-3, "Learning rate to use.")
flags.DEFINE_integer(
    "log_interval", 10, "In which intervals to display learning progress."
)
flags.DEFINE_integer("model_save_interval", 50, "Save model every so many epochs.")
flags.DEFINE_boolean("save_model", True, "Save the model after training.")
flags.DEFINE_boolean("big_net", False, "Use bigger net...")
flags.DEFINE_boolean("only_output", False, "Train only the last layer...")
flags.DEFINE_boolean("do_plot", False, "Do intermediate plots")
flags.DEFINE_integer("random_seed", 1234, "Random seed to use")


class LIFConvNet(torch.nn.Module):
    def __init__(
        self,
        input_features,
        seq_length,
        model="super",
        only_first_spike=False,
    ):
        super(LIFConvNet, self).__init__()
        self.constant_current_encoder = ConstantCurrentLIFEncoder(seq_length=seq_length)
        self.only_first_spike = only_first_spike
        self.input_features = input_features
        self.rsnn = ConvNet4(method=model)
        self.seq_length = seq_length

    def forward(self, x):
        batch_size = x.shape[0]
        x = self.constant_current_encoder(
            x.view(-1, self.input_features) * FLAGS.input_scale
        )
        if self.only_first_spike:
            # delete all spikes except for first
            zeros = torch.zeros_like(x.cpu()).detach().numpy()
            idxs = x.cpu().nonzero().detach().numpy()
            spike_counter = np.zeros((FLAGS.batch_size, 28 * 28))
            for t, batch, nrn in idxs:
                if spike_counter[batch, nrn] == 0:
                    zeros[t, batch, nrn] = 1
                    spike_counter[batch, nrn] += 1
            x = torch.from_numpy(zeros).to(x.device)

        x = x.reshape(self.seq_length, batch_size, 1, 28, 28)
        voltages = self.rsnn(x)
        m, _ = torch.max(voltages, 0)
        log_p_y = torch.nn.functional.log_softmax(m, dim=1)
        return log_p_y


def train(model, device, train_loader, optimizer, epoch, writer=None):
    model.train()
    losses = []

    batch_len = len(train_loader)
    step = batch_len * epoch

    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = torch.nn.functional.nll_loss(output, target)
        loss.backward()

        if FLAGS.clip_grad:
            torch.nn.utils.clip_grad_norm_(model.parameters(), FLAGS.grad_clip_value)

        optimizer.step()
        step += 1

        if batch_idx % FLAGS.log_interval == 0:
            logging.info(
                "Train Epoch: {}/{} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    epoch,
                    FLAGS.epochs,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                )
            )

        if step % FLAGS.log_interval == 0 and writer:
            _, argmax = torch.max(output, 1)
            accuracy = (target == argmax.squeeze()).float().mean()

            writer.add_scalar("Loss/train", loss.item(), step)
            writer.add_scalar("Accuracy/train", accuracy.item(), step)

            for tag, value in model.named_parameters():
                tag = tag.replace(".", "/")
                writer.add_histogram(tag, value.data.cpu().numpy(), step)
                writer.add_histogram(tag + "/grad", value.grad.data.cpu().numpy(), step)

        if FLAGS.do_plot and batch_idx % FLAGS.plot_interval == 0:
            ts = np.arange(0, FLAGS.seq_length)
            fig, axs = plt.subplots(4, 4, figsize=(15, 10), sharex=True, sharey=True)
            axs = axs.reshape(-1)  # flatten
            for nrn in range(10):
                one_trace = model.voltages.detach().cpu().numpy()[:, 0, nrn]
                fig.sca(axs[nrn])
                fig.plot(ts, one_trace)
            fig.xlabel("Time [s]")
            fig.ylabel("Membrane Potential")

            writer.add_figure("Voltages/output", fig, step)

        losses.append(loss.item())

    mean_loss = np.mean(losses)
    return losses, mean_loss


def test(model, device, test_loader, epoch, writer=None):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += torch.nn.functional.nll_loss(
                output, target, reduction="sum"
            ).item()  # sum up batch loss
            pred = output.argmax(
                dim=1, keepdim=True
            )  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    accuracy = 100.0 * correct / len(test_loader.dataset)
    logging.info(
        f"\nTest set {FLAGS.model}: Average loss: {test_loss:.4f}, \
            Accuracy: {correct}/{len(test_loader.dataset)} ({accuracy:.0f}%)\n"
    )
    if writer:
        writer.add_scalar("Loss/test", test_loss, epoch)
        writer.add_scalar("Accuracy/test", accuracy, epoch)

    return test_loss, accuracy


def save(path, epoch, model, optimizer, is_best=False):
    torch.save(
        {
            "epoch": epoch + 1,
            "model_state_dict": model.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
            "is_best": is_best,
        },
        path,
    )


def load(path, model, optimizer):
    checkpoint = torch.load(path)
    model.load_state_dict(checkpoint["model_state_dict"])
    optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
    model.train()
    return model, optimizer


def main(argv):
    try:
        from torch.utils.tensorboard import SummaryWriter

        writer = SummaryWriter()
    except ImportError:
        writer = None

    torch.manual_seed(FLAGS.random_seed)

    np.random.seed(FLAGS.random_seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(FLAGS.random_seed)
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.benchmark = True

    device = torch.device(FLAGS.device)

    kwargs = {"num_workers": 1, "pin_memory": True} if FLAGS.device == "cuda" else {}
    train_loader = torch.utils.data.DataLoader(
        torchvision.datasets.MNIST(
            root=".",
            train=True,
            download=True,
            transform=torchvision.transforms.Compose(
                [
                    # torchvision.transforms.
                    #    RandomCrop(size=[28,28], padding=4)
                    torchvision.transforms.ToTensor(),
                    torchvision.transforms.Normalize((0.1307,), (0.3081,)),
                ]
            ),
        ),
        batch_size=FLAGS.batch_size,
        shuffle=True,
        **kwargs,
    )
    test_loader = torch.utils.data.DataLoader(
        torchvision.datasets.MNIST(
            root=".",
            train=False,
            transform=torchvision.transforms.Compose(
                [
                    torchvision.transforms.ToTensor(),
                    torchvision.transforms.Normalize((0.1307,), (0.3081,)),
                ]
            ),
        ),
        batch_size=FLAGS.batch_size,
        **kwargs,
    )

    label = os.environ.get("SLURM_JOB_ID", str(uuid.uuid4()))
    if FLAGS.prefix:
        path = f"runs/mnist/{FLAGS.prefix}/{label}"
    else:
        path = f"runs/mnist/{label}"

    os.makedirs(path, exist_ok=True)
    os.chdir(path)
    FLAGS.append_flags_into_file("flags.txt")

    input_features = 28 * 28

    model = LIFConvNet(
        input_features,
        FLAGS.seq_length,
        model=FLAGS.model,
        only_first_spike=FLAGS.only_first_spike,
    ).to(device)

    if FLAGS.optimizer == "sgd":
        optimizer = torch.optim.SGD(model.parameters(), lr=FLAGS.learning_rate)
    elif FLAGS.optimizer == "adam":
        optimizer = torch.optim.Adam(model.parameters(), lr=FLAGS.learning_rate)

    if FLAGS.only_output:
        optimizer = torch.optim.Adam(model.out.parameters(), lr=FLAGS.learning_rate)

    training_losses = []
    mean_losses = []
    test_losses = []
    accuracies = []

    for epoch in range(FLAGS.epochs):
        training_loss, mean_loss = train(
            model, device, train_loader, optimizer, epoch, writer=writer
        )
        test_loss, accuracy = test(model, device, test_loader, epoch, writer=writer)

        training_losses += training_loss
        mean_losses.append(mean_loss)
        test_losses.append(test_loss)
        accuracies.append(accuracy)

        max_accuracy = np.max(np.array(accuracies))

        if (epoch % FLAGS.model_save_interval == 0) and FLAGS.save_model:
            model_path = f"mnist-{epoch}.pt"
            save(
                model_path,
                model=model,
                optimizer=optimizer,
                epoch=epoch,
                is_best=accuracy > max_accuracy,
            )

    np.save("training_losses.npy", np.array(training_losses))
    np.save("mean_losses.npy", np.array(mean_losses))
    np.save("test_losses.npy", np.array(test_losses))
    np.save("accuracies.npy", np.array(accuracies))
    model_path = "mnist-final.pt"
    save(
        model_path,
        epoch=epoch,
        model=model,
        optimizer=optimizer,
        is_best=accuracy > max_accuracy,
    )
    if writer:
        writer.close()


if __name__ == "__main__":
    app.run(main)



from typing import Optional, Tuple

import torch
import numpy as np

from ..functional.lsnn import (
    LSNNParameters,
    LSNNState,
    LSNNFeedForwardState,
    lsnn_step,
    lsnn_feed_forward_step,
)


class LSNNCell(torch.nn.Module):
    r"""Module that computes a single euler-integration step of a LSNN
    neuron-model. More specifically it implements one integration step of
    the following ODE

    .. math::
        \\begin{align*}
            \dot{v} &= 1/\\tau_{\\text{mem}} (v_{\\text{leak}} - v + i) \\\\
            \dot{i} &= -1/\\tau_{\\text{syn}} i \\\\
            \dot{b} &= -1/\\tau_{b} b
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\\text{th}} + b)

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            i &= i + w_{\\text{input}} z_{\\text{in}} \\\\
            i &= i + w_{\\text{rec}} z_{\\text{rec}} \\\\
            b &= b + \\beta z
        \end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the
    recurrent and input spikes respectively.

    Parameters:
        input_size (int): Size of the input. Also known as the number of input features.
        hidden_size (int): Size of the hidden state. Also known as the number of input features.
        p (LSNNParameters): parameters of the lsnn unit
        dt (float): Integration timestep to use
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LSNNParameters = LSNNParameters(),
        dt: float = 0.001,
    ):
        super(LSNNCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size)
        )
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LSNNState] = None
    ) -> Tuple[torch.Tensor, LSNNState]:
        if state is None:
            state = LSNNState(
                z=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                v=self.p.v_leak.detach(),
                i=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                b=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.v.requires_grad = True
        return lsnn_step(
            input_tensor,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )


class LSNNLayer(torch.nn.Module):
    r"""A Long short-term memory neuron module adapted from
        https://arxiv.org/abs/1803.09574

    Usage:
      >>> from norse.torch.module import LSNNLayer, LSNNCell
      >>> layer = LSNNLayer(LSNNCell, 2, 10)    // LSNNCell of shape 2 -> 10
      >>> data  = torch.zeros(2, 5, 2)          // Arbitrary data
      >>> output, state = layer.forward(data)

    Parameters:
      cell (torch.nn.Module): the underling neuron module, uninitialized
      *cell_args: variable length input arguments for the underlying cell
                  constructor
      **cell_kwargs: variable length key-value arguments for the underlying cell constructor
    """

    def __init__(self, cell, *cell_args, **cell_kwargs):
        super(LSNNLayer, self).__init__()
        self.cell = cell(*cell_args, **cell_kwargs)

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LSNNState] = None
    ) -> Tuple[torch.Tensor, LSNNState]:
        """
        Takes one step in the LSNN layer by simulating the layer for a number of timesteps.
        Since the layer is recurrent, each simulation state (LSNNState) is used as the input for the next step.

        The function expects inputs in the shape (simulation time steps, batch size, ...).

        Parameters:
            input_tensor (torch.Tensor): Input tensor with timesteps in the first dimension
            state (Optional[LSNNState]): The input LSNN state. Defaults to None on the first timestep

        Returns:
            A tuple of 1) spikes from each timestep and 2) the LSNNState from the last timestep.
        """
        inputs = input_tensor.unbind(0)
        outputs = []  # torch.jit.annotate(List[torch.Tensor], [])
        for input_step in inputs:
            out, state = self.cell(input_step, state)
            outputs += [out]
        # pytype: disable=bad-return-type
        return torch.stack(outputs), state
        # pytype: enable=bad-return-type


class LSNNFeedForwardCell(torch.nn.Module):
    r"""Euler integration cell for LIF Neuron with threshold adaptation.
    More specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} - v + i) \\
            \dot{i} &= -1/\tau_{\\text{syn}} i \\
            \dot{b} &= -1/\tau_{b} b
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}} + b)

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + \text{input} \\
            b &= b + \beta z
        \end{align*}

    Parameters:
        s (LSNNFeedForwardState): current state of the lsnn unit
        p (LSNNParameters): parameters of the lsnn unit
        dt (float): Integration timestep to use
    """

    def __init__(self, p: LSNNParameters = LSNNParameters(), dt: float = 0.001):
        super(LSNNFeedForwardCell, self).__init__()
        self.p = p
        self.dt = dt

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LSNNFeedForwardState] = None
    ) -> Tuple[torch.Tensor, LSNNFeedForwardState]:
        if state is None:
            state = LSNNFeedForwardState(
                v=self.p.v_leak.detach(),
                i=torch.zeros(
                    *input_tensor.shape,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                b=torch.zeros(
                    *input_tensor.shape,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.v.requires_grad = True
        return lsnn_feed_forward_step(input_tensor, state, p=self.p, dt=self.dt)



from typing import NamedTuple

import torch


class STDPSensorParameters(NamedTuple):
    """Parameters of an STDP sensor as it is used for event driven
    plasticity rules.

    Parameters:
        eta_p (torch.Tensor): correlation state
        eta_m (torch.Tensor): anti correlation state
        tau_ac_inv (torch.Tensor): anti-correlation sensor time constant
        tau_c_inv (torch.Tensor): correlation sensor time constant
    """

    eta_p: torch.Tensor = torch.as_tensor(1.0)
    eta_m: torch.Tensor = torch.as_tensor(1.0)
    tau_ac_inv: torch.Tensor = torch.as_tensor(1.0 / 100e-3)
    tau_c_inv: torch.Tensor = torch.as_tensor(1.0 / 100e-3)


class STDPSensorState(NamedTuple):
    """State of an event driven STDP sensor.

    Parameters:
        a_pre (torch.Tensor): presynaptic STDP sensor state.
        a_post (torch.Tensor): postsynaptic STDP sensor state.
    """

    a_pre: torch.Tensor
    a_post: torch.Tensor


def stdp_sensor_step(
    z_pre: torch.Tensor,
    z_post: torch.Tensor,
    state: STDPSensorState,
    p: STDPSensorParameters = STDPSensorParameters(),
    dt: float = 0.001,
) -> STDPSensorState:
    """Event driven STDP rule.

    Parameters:
        z_pre (torch.Tensor): pre-synaptic spikes
        z_post (torch.Tensor): post-synaptic spikes
        s (STDPSensorState): state of the STDP sensor
        p (STDPSensorParameters): STDP sensor parameters
        dt (float): integration time step
    """
    da_pre = p.tau_c_inv * (-state.a_pre)
    a_pre_decayed = state.a_pre + dt * da_pre

    da_post = p.tau_c_inv * (-state.a_post)
    a_post_decayed = state.a_post + dt * da_post

    a_pre_new = a_pre_decayed + z_pre * p.eta_p
    a_post_new = a_post_decayed + z_post * p.eta_m

    return STDPSensorState(a_pre_new, a_post_new)



"""
Stateless encoding functionality for Norse, offering different ways to convert numerical
inputs to the spiking domain. Note that some functions, like `population_encode` does not return spikes,
but rather numerical values that will have to be converted into spikes via, for instance, the poisson encoder.
"""

from typing import Callable, Union

import torch

from .lif import lif_current_encoder, LIFParameters


def constant_current_lif_encode(
    input_current: torch.Tensor,
    seq_length: int,
    p: LIFParameters = LIFParameters(),
    dt: float = 0.001,
) -> torch.Tensor:
    """
    Encodes input currents as fixed (constant) voltage currents, and simulates the spikes that
    occur during a number of timesteps/iterations (seq_length).

    Example:
        >>> data = torch.as_tensor([2, 4, 8, 16])
        >>> seq_length = 2 # Simulate two iterations
        >>> constant_current_lif_encode(data, seq_length)
         # State in terms of membrane voltage
        (tensor([[0.2000, 0.4000, 0.8000, 0.0000],
                 [0.3800, 0.7600, 0.0000, 0.0000]]),
         # Spikes for each iteration
         tensor([[0., 0., 0., 1.],
                 [0., 0., 1., 1.]]))

    Parameters:
        input_current (torch.Tensor): The input tensor, representing LIF current
        seq_length (int): The number of iterations to simulate
        p (LIFParameters): Initial neuron parameters.
        dt (float): Time delta between simulation steps

    Returns:
        A tensor with an extra dimension of size `seq_length` containing spikes (1) or no spikes (0).
    """
    v = torch.zeros(*input_current.shape, device=input_current.device)
    z = torch.zeros(*input_current.shape, device=input_current.device)
    spikes = torch.zeros(seq_length, *input_current.shape, device=input_current.device)

    for ts in range(seq_length):
        z, v = lif_current_encoder(input_current=input_current, voltage=v, p=p, dt=dt)
        spikes[ts] = z
    return spikes


def gaussian_rbf(tensor: torch.Tensor, sigma: float = 1):
    """
    A `gaussian radial basis kernel <https://en.wikipedia.org/wiki/Radial_basis_function_kernel>`_
    that calculates the radial basis given a distance value (distance between :math:`x` and a data
    value :math:`x'`, or :math:`\\|\\mathbf{x} - \\mathbf{x'}\\|^2` below).

    .. math::
        K(\\mathbf{x}, \\mathbf{x'}) = \\exp\\left(- \\frac{\\|\\mathbf{x} - \\mathbf{x'}\\|^2}{2\\sigma^2}\\right)

    Parameters:
        tensor (torch.Tensor): The tensor containing distance values to convert to radial bases
        sigma (float): The spread of the gaussian distribution. Defaults to 1.
    """
    return torch.exp(-tensor / (2 * sigma ** 2))


def euclidean_distance(x, y):
    """
    Simple euclidean distance metric.
    """
    return (x - y).pow(2)


def population_encode(
    input_values: torch.Tensor,
    out_features: int,
    scale: Union[int, torch.Tensor] = None,
    kernel: Callable[[torch.Tensor], torch.Tensor] = gaussian_rbf,
    distance_function: Callable[
        [torch.Tensor, torch.Tensor], torch.Tensor
    ] = euclidean_distance,
) -> torch.Tensor:
    """
    Encodes a set of input values into population codes, such that each singular input value is represented by
    a list of numbers (typically calculated by a radial basis kernel), whose length is equal to the out_features.

    Population encoding can be visualised by imagining a number of neurons in a list, whose activity increases
    if a number gets close to its "receptive field".

    .. figure:: https://upload.wikimedia.org/wikipedia/commons/thumb/a/a1/PopulationCode.svg/1920px-PopulationCode.svg.png

        Gaussian curves representing different neuron "receptive fields". Image credit: `Andrew K. Richardson`_.

    .. _Andrew K. Richardson: https://commons.wikimedia.org/wiki/File:PopulationCode.svg

    Example:
        >>> data = torch.as_tensor([0, 0.5, 1])
        >>> out_features = 3
        >>> pop_encoded = population_encode(data, out_features)
        tensor([[1.0000, 0.8825, 0.6065],
                [0.8825, 1.0000, 0.8825],
                [0.6065, 0.8825, 1.0000]])
        >>> spikes = poisson_encode(pop_encoded, 1).squeeze() # Convert to spikes

    Parameters:
        input_values (torch.Tensor): The input data as numerical values to be encoded to population codes
        out_features (int): The number of output *per* input value
        scale (torch.Tensor): The scaling factor for the kernels. Defaults to the maximum value of the input.
                              Can also be set for each individual sample.
        kernel: A function that takes two inputs and returns a tensor. The two inputs represent the center value
                (which changes for each index in the output tensor) and the actual data value to encode respectively.z
                Defaults to gaussian radial basis kernel function.
        distance_function: A function that calculates the distance between two numbers. Defaults to euclidean.

    Returns:
        A tensor with an extra dimension of size `seq_length` containing population encoded values of the input stimulus.
        Note: An extra step is required to convert the values to spikes, see above.
    """
    # Thanks to: https://github.com/JeremyLinux/PyTorch-Radial-Basis-Function-Layer/blob/master/Torch%20RBF/torch_rbf.py
    size = (input_values.size(0), out_features) + input_values.size()[1:]
    if not scale:
        scale = input_values.max()
    centres = torch.linspace(0, scale, out_features).expand(size)
    x = input_values.unsqueeze(1).expand(size)
    distances = distance_function(x, centres) * scale
    return kernel(distances)


def poisson_encode(
    input_values: torch.Tensor,
    seq_length: int,
    f_max: float = 100,
    dt: float = 0.001,
) -> torch.Tensor:
    """
    Encodes a tensor of input values, which are assumed to be in the
    range [0,1] into a tensor of one dimension higher of binary values,
    which represent input spikes.

    See for example https://www.cns.nyu.edu/~david/handouts/poisson.pdf.

    Parameters:
        input_values (torch.Tensor): Input data tensor with values assumed to be in the interval [0,1].
        sequence_length (int): Number of time steps in the resulting spike train.
        f_max (float): Maximal frequency (in Hertz) which will be emitted.
        dt (float): Integration time step (should coincide with the integration time step used in the model)

    Returns:
        A tensor with an extra dimension of size `seq_length` containing spikes (1) or no spikes (0).
    """
    return (
        torch.rand(seq_length, *input_values.shape, device=input_values.device).float()
        < dt * f_max * input_values
    ).float()


def signed_poisson_encode(
    input_values: torch.Tensor, seq_length: int, f_max: float = 100, dt: float = 0.001
) -> torch.Tensor:
    """
    Encodes a tensor of input values, which are assumed to be in the
    range [-1,1] into a tensor of one dimension higher of binary values,
    which represent input spikes.

    Parameters:
        input_values (torch.Tensor): Input data tensor with values assumed to be in the interval [-1,1].
        sequence_length (int): Number of time steps in the resulting spike train.
        f_max (float): Maximal frequency (in Hertz) which will be emitted.
        dt (float): Integration time step (should coincide with the integration time step used in the model)

    Returns:
        A tensor with an extra dimension of size `seq_length` containing spikes (1) or no spikes (0).
    """
    return (
        torch.sign(input_values)
        * (
            torch.rand(seq_length, *input_values.shape).float()
            < dt * f_max * torch.abs(input_values)
        ).float()
    )


def spike_latency_lif_encode(
    input_current: torch.Tensor,
    seq_length: int,
    p: LIFParameters = LIFParameters(),
    dt=0.001,
) -> torch.Tensor:
    """Encodes an input value by the time the first spike occurs.
    Similar to the ConstantCurrentLIFEncoder, but the LIF can be
    thought to have an infinite refractory period.

    Parameters:
        input_current (torch.Tensor): Input current to encode (needs to be positive).
        sequence_length (int): Number of time steps in the resulting spike train.
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Integration time step (should coincide with the integration time step used in the model)
    """
    voltage = torch.zeros_like(input_current)
    z = torch.zeros_like(input_current)
    mask = torch.zeros_like(input_current).byte()
    spikes = []

    for _ in range(seq_length):
        z, voltage = lif_current_encoder(
            input_current=input_current, voltage=voltage, p=p, dt=dt
        )
        spikes += [torch.where(mask, torch.zeros_like(z), z)]
        mask[z.bool()] = 1

    return torch.stack(spikes)


def spike_latency_encode(input_spikes: torch.Tensor) -> torch.Tensor:
    """
    For all neurons, remove all but the first spike. This encoding basically measures the time it takes for a
    neuron to spike *first*. Assuming that the inputs are constant, this makes sense in that strong inputs spikes
    fast.

    See `R. Van Rullen & S. J. Thorpe (2001): Rate Coding Versus Temporal Order Coding: What the Retinal Ganglion Cells Tell the Visual Cortex <https://doi.org/10.1162/08997660152002852>`_.

    Spikes are identified by their unique position within each sequence.

    Example:
        >>> data = torch.as_tensor([[0, 1, 1], [1, 1, 1]])
        >>> spike_latency_encode(data)
        tensor([[0, 1, 1],
                [1, 0, 0]])

    Parameters:
        input_spikes (torch.Tensor): A tensor of input spikes, assumed to be at least 2D (sequences, ...)

    Returns:
        A tensor where the first spike (1) is retained in the sequence
    """
    mask = torch.zeros(input_spikes.size()[1:], device=input_spikes.device).byte()
    spikes = []
    zero_spikes = torch.zeros_like(input_spikes[0])
    for index in range(input_spikes.shape[0]):
        spikes.append(torch.where(mask, zero_spikes, input_spikes[index]))
        mask[input_spikes[index].bool()] = 1
    return torch.stack(spikes)



import torch

from norse.torch.module.lif import LIFFeedForwardCell
from norse.torch.module.regularization import RegularizationCell


def test_regularization_module():
    cell = LIFFeedForwardCell()
    r = RegularizationCell()  # Defaults to spike counting
    data = torch.ones(5, 2) + 10  # Batch size of 5
    z, s = cell(data)
    z, rs = r(z, s)
    assert z.shape == (5, 2)
    assert rs == 0
    z, s = cell(data, s)
    z, rs = r(z, s)
    assert rs == 10
    assert r.state == 10



import torch

from norse.torch.module.lif_ex import (
    LIFExCell,
    LIFExLayer,
    LIFExState,
    LIFExFeedForwardCell,
    LIFExFeedForwardState,
)


def test_lif_ex_cell():
    cell = LIFExCell(2, 4)
    data = torch.randn(5, 2)
    out, _ = cell(data)

    assert out.shape == (5, 4)


def test_lif_ex_cell_backward():
    cell = LIFExCell(2, 4)
    data = torch.randn(5, 2)
    out, _ = cell(data)
    out.sum().backward()


def test_lif_ex_layer():
    layer = LIFExLayer(2, 4)
    data = torch.randn(10, 5, 2)
    out, _ = layer(data)

    assert out.shape == (10, 5, 4)


def test_lif_ex_layer_state():
    layer = LIFExLayer(2, 4)
    input_tensor = torch.randn(10, 5, 2)

    state = LIFExState(
        z=torch.zeros(
            (input_tensor.shape[1], layer.cell.hidden_size),
            dtype=input_tensor.dtype,
            device=input_tensor.device,
        ),
        v=layer.cell.p.v_leak,
        i=torch.zeros(
            input_tensor.shape[1],
            layer.cell.hidden_size,
            device=input_tensor.device,
            dtype=input_tensor.dtype,
        ),
    )
    out, _ = layer(input_tensor, state)

    assert out.shape == (10, 5, 4)


def test_lif_ex_feedforward_cell():
    layer = LIFExFeedForwardCell()
    data = torch.randn(5, 2, 4)
    out, _ = layer(data)

    assert out.shape == (5, 2, 4)


def test_lif_ex_feedforward_cell_state():
    layer = LIFExFeedForwardCell()
    input_tensor = torch.randn(5, 2, 4)
    state = LIFExFeedForwardState(
        v=layer.p.v_leak,
        i=torch.zeros(
            *input_tensor.shape, device=input_tensor.device, dtype=input_tensor.dtype
        ),
    )
    out, _ = layer(input_tensor, state)

    assert out.shape == (5, 2, 4)


def test_lif_ex_feedforward_cell_backward():
    layer = LIFExFeedForwardCell()
    data = torch.randn(5, 2, 4)
    out, _ = layer(data)
    out.sum().backward()


def test_lif_ex_feedforward_cell_repr():
    model = LIFExFeedForwardCell()
    assert (
        str(model)
        == "LIFExFeedForwardCell(p=LIFExParameters(delta_T=tensor(0.5000), tau_syn_inv=tensor(200.), tau_mem_inv=tensor(100.), v_leak=tensor(0.), v_th=tensor(1.), v_reset=tensor(0.), method='super', alpha=100.0), dt=0.001)"
    )


def test_lif_ex_cell_repr():
    model = LIFExCell(2, 4)
    assert (
        str(model)
        == "LIFExCell(2, 4, p=LIFExParameters(delta_T=tensor(0.5000), tau_syn_inv=tensor(200.), tau_mem_inv=tensor(100.), v_leak=tensor(0.), v_th=tensor(1.), v_reset=tensor(0.), method='super', alpha=100.0), dt=0.001)"
    )



import torch

import numpy as np
from typing import Optional, Tuple

from ..functional.lif import LIFState, LIFParameters
from ..functional.lif_mc import lif_mc_step


class LIFMCCell(torch.nn.Module):
    r"""Computes a single euler-integration step of a LIF multi-compartment
    neuron-model.

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} (v_{\text{leak}} \
            - g_{\text{coupling}} v + i) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + w_{\text{input}} z_{\text{in}} \\
            i &= i + w_{\text{rec}} z_{\text{rec}}
        \end{align*}

    where :math:`z_{\text{rec}}` and :math:`z_{\text{in}}` are the
    recurrent and input spikes respectively.


    Parameters:
        input_size (int): Size of the input. Also known as the number of input features.
        hidden_size (int): Size of the hidden state. Also known as the number of input features.
        g_coupling (torch.Tensor): conductances between the neuron compartments
        p (LIFParameters): neuron parameters
        dt (float): Integration timestep to use
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: LIFParameters = LIFParameters(),
        dt: float = 0.001,
    ):
        super(LIFMCCell, self).__init__()

        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.g_coupling = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[LIFState] = None
    ) -> Tuple[torch.Tensor, LIFState]:
        if state is None:
            state = LIFState(
                z=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                v=self.p.v_leak.detach()
                * torch.ones(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                i=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
        return lif_mc_step(
            input_tensor,
            state,
            self.input_weights,
            self.recurrent_weights,
            self.g_coupling,
            p=self.p,
            dt=self.dt,
        )



"""Norse is a library for doing deep learning with spiking neural networks.

This package contains modules that extends PyTorch with spiking neural
network functionality.
"""

from .functional import *
from .models import *
from .module import *



from typing import NamedTuple, Tuple

import torch
import torch.jit

from norse.torch.functional.threshold import threshold


class LIFExParameters(NamedTuple):
    """Parametrization of an Exponential Leaky Integrate and Fire neuron

    Parameters:
        delta_T (torch.Tensor): sharpness or speed of the exponential growth in mV
        tau_syn_inv (torch.Tensor): inverse synaptic time
                                    constant (:math:`1/\\tau_\\text{syn}`) in 1/ms
        tau_mem_inv (torch.Tensor): inverse membrane time
                                    constant (:math:`1/\\tau_\\text{mem}`) in 1/ms
        v_leak (torch.Tensor): leak potential in mV
        v_th (torch.Tensor): threshold potential in mV
        v_reset (torch.Tensor): reset potential in mV
        method (str): method to determine the spike threshold
                      (relevant for surrogate gradients)
        alpha (float): hyper parameter to use in surrogate gradient computation
    """

    delta_T: torch.Tensor = torch.as_tensor(0.5)
    tau_syn_inv: torch.Tensor = torch.as_tensor(1.0 / 5e-3)
    tau_mem_inv: torch.Tensor = torch.as_tensor(1.0 / 1e-2)
    v_leak: torch.Tensor = torch.as_tensor(0.0)
    v_th: torch.Tensor = torch.as_tensor(1.0)
    v_reset: torch.Tensor = torch.as_tensor(0.0)
    method: str = "super"
    alpha: float = 100.0


class LIFExState(NamedTuple):
    """State of a LIFEx neuron

    Parameters:
        z (torch.Tensor): recurrent spikes
        v (torch.Tensor): membrane potential
        i (torch.Tensor): synaptic input current
    """

    z: torch.Tensor
    v: torch.Tensor
    i: torch.Tensor


class LIFExFeedForwardState(NamedTuple):
    """State of a feed forward LIFEx neuron

    Parameters:
        v (torch.Tensor): membrane potential
        i (torch.Tensor): synaptic input current
    """

    v: torch.Tensor
    i: torch.Tensor


def lif_ex_step(
    input_tensor: torch.Tensor,
    state: LIFExState,
    input_weights: torch.Tensor,
    recurrent_weights: torch.Tensor,
    p: LIFExParameters = LIFExParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFExState]:
    r"""Computes a single euler-integration step of an exponential LIF neuron-model
    adapted from https://neuronaldynamics.epfl.ch/online/Ch5.S2.html. More
    specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} \left(v_{\text{leak}} - v + i + \Delta_T exp\left({{v - v_{\text{th}}} \over {\Delta_T}}\right)\right) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + w_{\text{input}} z_{\text{in}} \\
            i &= i + w_{\text{rec}} z_{\text{rec}}
        \end{align*}

    where :math:`z_{\text{rec}}` and :math:`z_{\text{in}}` are the recurrent
    and input spikes respectively.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        s (LIFExState): current state of the LIF neuron
        input_weights (torch.Tensor): synaptic weights for incoming spikes
        recurrent_weights (torch.Tensor): synaptic weights for recurrent spikes
        p (LIFExParameters): parameters of a leaky integrate and fire neuron
        dt (float): Integration timestep to use
    """
    # compute voltage updates
    dv_leak = p.v_leak - state.v
    dv_exp = p.delta_T * torch.exp((state.v - p.v_th) / p.delta_T)
    dv = dt * p.tau_mem_inv * (dv_leak + dv_exp + state.i)
    v_decayed = state.v + dv

    # compute current updates
    di = -dt * p.tau_syn_inv * state.i
    i_decayed = state.i + di

    # compute new spikes
    z_new = threshold(v_decayed - p.v_th, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute current jumps
    i_new = (
        i_decayed
        + torch.nn.functional.linear(input_tensor, input_weights)
        + torch.nn.functional.linear(state.z, recurrent_weights)
    )

    return z_new, LIFExState(z_new, v_new, i_new)


def lif_ex_feed_forward_step(
    input_tensor: torch.Tensor,
    state: LIFExFeedForwardState = LIFExFeedForwardState(0, 0),
    p: LIFExParameters = LIFExParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, LIFExFeedForwardState]:
    r"""Computes a single euler-integration step of an exponential LIF neuron-model
    adapted from https://neuronaldynamics.epfl.ch/online/Ch5.S2.html.
    It takes as input the input current as generated by an arbitrary torch
    module or function. More specifically it implements one integration step of
    the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} \left(v_{\text{leak}} - v + i + \Delta_T exp\left({{v - v_{\text{th}}} \over {\Delta_T}}\right)\right) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    together with the jump condition

    .. math::
        z = \Theta(v - v_{\text{th}})

    and transition equations

    .. math::
        \begin{align*}
            v &= (1-z) v + z v_{\text{reset}} \\
            i &= i + i_{\text{in}}
        \end{align*}

    where :math:`i_{\text{in}}` is meant to be the result of applying an
    arbitrary pytorch module (such as a convolution) to input spikes.

    Parameters:
        input_tensor (torch.Tensor): the input spikes at the current time step
        state (LIFExFeedForwardState): current state of the LIF neuron
        p (LIFExParameters): parameters of a leaky integrate and fire neuron
        dt (float): Integration timestep to use
    """
    # compute voltage updates
    dv_leak = p.v_leak - state.v
    dv_exp = p.delta_T * torch.exp((state.v - p.v_th) / p.delta_T)
    dv = dt * p.tau_mem_inv * (dv_leak + dv_exp + state.i)
    v_decayed = state.v + dv

    # compute current updates
    di = -dt * p.tau_syn_inv * state.i
    i_decayed = state.i + di

    # compute new spikes
    z_new = threshold(v_decayed - p.v_th, p.method, p.alpha)
    # compute reset
    v_new = (1 - z_new) * v_decayed + z_new * p.v_reset
    # compute current jumps
    i_new = i_decayed + input_tensor

    return z_new, LIFExFeedForwardState(v_new, i_new)


def lif_ex_current_encoder(
    input_current: torch.Tensor,
    voltage: torch.Tensor,
    p: LIFExParameters = LIFExParameters(),
    dt: float = 0.001,
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""Computes a single euler-integration step of a leaky integrator
    adapted from https://neuronaldynamics.epfl.ch/online/Ch5.S2.html. More
    specifically it implements one integration step of the following ODE

    .. math::
        \begin{align*}
            \dot{v} &= 1/\tau_{\text{mem}} \left(v_{\text{leak}} - v + i + \Delta_T exp\left({{v - v_{\text{th}}} \over {\Delta_T}}\right)\right) \\
            \dot{i} &= -1/\tau_{\text{syn}} i
        \end{align*}

    Parameters:
        input (torch.Tensor): the input current at the current time step
        voltage (torch.Tensor): current state of the LIFEx neuron
        p (LIFExParameters): parameters of a leaky integrate and fire neuron
        dt (float): Integration timestep to use
    """
    dv_leak = p.v_leak - voltage
    dv_exp = p.delta_T * torch.exp((voltage - p.v_th) / p.delta_T)
    dv = dt * p.tau_mem_inv * (dv_leak + dv_exp + input_current)
    voltage = voltage + dv
    z = threshold(voltage - p.v_th, p.method, p.alpha)

    voltage = voltage - z * (voltage - p.v_reset)
    return z, voltage



from typing import NamedTuple, Tuple

import torch


class TsodyksMakramState(NamedTuple):
    """State of the Tsodyks-Makram Model, note
    that we are tracking the input current state separately.

    Parameters:
        u (torch.Tensor): utilization parameter.
        x (torch.Tensor): fraction of remaining available resources.
    """

    u: torch.Tensor
    x: torch.Tensor


class TsodyksMakramParameters(NamedTuple):
    """Parameters of the Tsodyks-Makram Model

    Parameters:
        tau_f_inv (float): facilitation time constant (in ms).
        tau_s_inv (float): synaptic time constant (in ms).
        tau_d_inv (float): depressing time constant (in ms).
        U (float): size of the jump in the utilization variable.
    """

    tau_f_inv: float = 1 / (50.0e-3)
    tau_s_inv: float = 1 / (20.0e-3)
    tau_d_inv: float = 1 / (750.0e-3)
    U: float = 0.45


def stp_step(
    z: torch.Tensor,
    s: TsodyksMakramState,
    p: TsodyksMakramParameters,
    dt: float = 0.001,
) -> Tuple[torch.Tensor, TsodyksMakramState]:
    """Euler integration step for Tsodyks Makram model of STP.

    Reference: http://www.scholarpedia.org/article/Short-term_synaptic_plasticity

    Parameters:
        z (torch.Tensor): Input spikes
        s (TsodyksMakramState): State of the Tsodyks-Makram model
        p (TsodyksMakramParameters): Parameters of the Tsodyks-Makram model
        dt (float): Euler integration timestep
    """
    du = -p.tau_f_inv * s.u
    u_p = s.u + dt * du + p.U * (1 - s.u) * z

    dx = p.tau_d_inv * (1 - s.x)
    x_p = s.x + dt * dx - u_p * s.x * z

    return u_p * s.x * z, TsodyksMakramState(u=u_p, x=x_p)



import torch

from norse.torch.functional.lif import LIFState
from norse.torch.functional.lif_refrac import LIFRefracState
from norse.torch.module.lif_mc_refrac import LIFMCRefracCell


def test_lif_mc_cell():
    cell = LIFMCRefracCell(2, 4)
    data = torch.randn(5, 2)
    out, s = cell(data)
    assert s.lif.v.shape == (5, 4)
    assert s.lif.i.shape == (5, 4)
    assert s.lif.z.shape == (5, 4)
    assert out.shape == (5, 4)


def test_lif_mc_refrac_state():
    cell = LIFMCRefracCell(2, 4)
    input_tensor = torch.randn(5, 2)
    state = LIFRefracState(
        lif=LIFState(
            z=torch.zeros(
                input_tensor.shape[0],
                cell.hidden_size,
            ),
            v=cell.p.lif.v_leak
            * torch.ones(
                input_tensor.shape[0],
                cell.hidden_size,
            ),
            i=torch.zeros(
                input_tensor.shape[0],
                cell.hidden_size,
            ),
        ),
        rho=torch.zeros(
            input_tensor.shape[0],
            cell.hidden_size,
        ),
    )

    out, s = cell(input_tensor, state)
    assert s.lif.v.shape == (5, 4)
    assert s.lif.i.shape == (5, 4)
    assert s.lif.z.shape == (5, 4)
    assert out.shape == (5, 4)


def test_lif_mc_cell_backward():
    cell = LIFMCRefracCell(2, 4)
    data = torch.randn(5, 2)
    out, _ = cell(data)
    out.sum().backward()



import torch

from typing import Optional, Tuple
import numpy as np

from norse.torch.functional.coba_lif import CobaLIFParameters, CobaLIFState
from norse.torch.functional.coba_lif import coba_lif_step


class CobaLIFCell(torch.nn.Module):
    """Module that computes a single euler-integration step of a conductance based
    LIF neuron-model. More specifically it implements one integration step of
    the following ODE

    .. math::
        \\begin{align*}
            \\dot{v} &= 1/c_{\\text{mem}} (g_l (v_{\\text{leak}} - v) \
              + g_e (E_{\\text{rev_e}} - v) + g_i (E_{\\text{rev_i}} - v)) \\\\
            \\dot{g_e} &= -1/\\tau_{\\text{syn}} g_e \\\\
            \\dot{g_i} &= -1/\\tau_{\\text{syn}} g_i
        \\end{align*}

    together with the jump condition

    .. math::
        z = \\Theta(v - v_{\\text{th}})

    and transition equations

    .. math::
        \\begin{align*}
            v &= (1-z) v + z v_{\\text{reset}} \\\\
            g_e &= g_e + \\text{relu}(w_{\\text{input}}) z_{\\text{in}} \\\\
            g_e &= g_e + \\text{relu}(w_{\\text{rec}}) z_{\\text{rec}} \\\\
            g_i &= g_i + \\text{relu}(-w_{\\text{input}}) z_{\\text{in}} \\\\
            g_i &= g_i + \\text{relu}(-w_{\\text{rec}}) z_{\\text{rec}} \\\\
        \\end{align*}

    where :math:`z_{\\text{rec}}` and :math:`z_{\\text{in}}` are the recurrent
    and input spikes respectively.

    Parameters:
        input_size (int): Size of the input.
        hidden_size (int): Size of the hidden state.
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Time step to use.

    Examples:

        >>> batch_size = 16
        >>> lif = CobaLIFCell(10, 20)
        >>> input = torch.randn(batch_size, 10)
        >>> output, s0 = lif(input)
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        p: CobaLIFParameters = CobaLIFParameters(),
        dt: float = 0.001,
    ):
        super(CobaLIFCell, self).__init__()
        self.input_weights = torch.nn.Parameter(
            torch.randn(hidden_size, input_size) / np.sqrt(input_size)
        )
        self.recurrent_weights = torch.nn.Parameter(
            torch.randn(hidden_size, hidden_size) / np.sqrt(hidden_size)
        )
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.p = p
        self.dt = dt

    def forward(
        self, input_tensor: torch.Tensor, state: Optional[CobaLIFState] = None
    ) -> Tuple[torch.Tensor, CobaLIFState]:
        if state is None:
            state = CobaLIFState(
                z=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                v=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                g_e=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
                g_i=torch.zeros(
                    input_tensor.shape[0],
                    self.hidden_size,
                    device=input_tensor.device,
                    dtype=input_tensor.dtype,
                ),
            )
            state.v.requires_grad = True
        return coba_lif_step(
            input_tensor,
            state,
            self.input_weights,
            self.recurrent_weights,
            p=self.p,
            dt=self.dt,
        )



# Parts of this code were adapted from the pytorch example at
# https://github.com/pytorch/examples/blob/master/reinforcement_learning/reinforce.py
# which is licensed under the license found in LICENSE.cartpole

import torch
import numpy as np
from absl import app
from absl import flags
from absl import logging
import random
import os

# pytype: disable=import-error
import gym

# pytype: enable=import-error

from norse.torch.functional.lif import LIFParameters
from norse.torch.module.encode import ConstantCurrentLIFEncoder
from norse.torch.module.lif import LIFCell
from norse.torch.module.lsnn import LSNNCell, LSNNParameters
from norse.torch.module.leaky_integrator import LICell

FLAGS = flags.FLAGS
flags.DEFINE_enum("device", "cpu", ["cpu", "cuda"], "Device to use by pytorch.")
flags.DEFINE_integer("episodes", 100, "Number of training trials.")
flags.DEFINE_float("learning_rate", 1e-3, "Learning rate to use.")
flags.DEFINE_float("gamma", 0.99, "discount factor to use")
flags.DEFINE_integer(
    "log_interval", 10, "In which intervals to display learning progress."
)
flags.DEFINE_enum("model", "super", ["super"], "Model to use for training.")
flags.DEFINE_enum("policy", "snn", ["snn", "lsnn", "ann"], "Select policy to use.")
flags.DEFINE_boolean("render", False, "Render the environment")
flags.DEFINE_string("environment", "CartPole-v1", "Gym environment to use.")
flags.DEFINE_integer("random_seed", 1234, "Random seed to use")


class ANNPolicy(torch.nn.Module):
    def __init__(self):
        super(ANNPolicy, self).__init__()
        self.state_space = 4
        self.action_space = 2
        self.l1 = torch.nn.Linear(self.state_space, 128, bias=False)
        self.l2 = torch.nn.Linear(128, self.action_space, bias=False)
        self.dropout = torch.nn.Dropout(p=0.6)

        self.saved_log_probs = []
        self.rewards = []

    def forward(self, x):
        x = self.l1(x)
        x = self.dropout(x)
        x = torch.nn.functional.relu(x)
        x = self.l2(x)
        x = torch.nn.functional.softmax(x)
        return x


class Policy(torch.nn.Module):
    def __init__(self):
        super(Policy, self).__init__()
        self.state_dim = 4
        self.input_features = 16
        self.hidden_features = 128
        self.output_features = 2
        self.constant_current_encoder = ConstantCurrentLIFEncoder(40)
        self.lif = LIFCell(
            2 * self.state_dim,
            self.hidden_features,
            p=LIFParameters(method="super", alpha=100.0),
        )
        self.dropout = torch.nn.Dropout(p=0.5)
        self.readout = LICell(self.hidden_features, self.output_features)

        self.saved_log_probs = []
        self.rewards = []

    def forward(self, x):
        scale = 50
        x_pos = self.constant_current_encoder(torch.nn.functional.relu(scale * x))
        x_neg = self.constant_current_encoder(torch.nn.functional.relu(-scale * x))
        x = torch.cat([x_pos, x_neg], dim=2)

        seq_length, batch_size, _ = x.shape

        voltages = torch.zeros(
            seq_length, batch_size, self.output_features, device=x.device
        )

        s1 = so = None
        # sequential integration loop
        for ts in range(seq_length):
            z1, s1 = self.lif(x[ts, :, :], s1)
            z1 = self.dropout(z1)
            vo, so = self.readout(z1, so)
            voltages[ts, :, :] = vo

        m, _ = torch.max(voltages, 0)
        p_y = torch.nn.functional.softmax(m, dim=1)
        return p_y


class LSNNPolicy(torch.nn.Module):
    def __init__(self, model="super"):
        super(LSNNPolicy, self).__init__()
        self.state_dim = 4
        self.input_features = 16
        self.hidden_features = 128
        self.output_features = 2
        # self.affine1 = torch.nn.Linear(self.state_dim, self.input_features)
        self.constant_current_encoder = ConstantCurrentLIFEncoder(40)
        self.lif_layer = LSNNCell(
            2 * self.state_dim,
            self.hidden_features,
            p=LSNNParameters(method=model, alpha=100.0),
        )
        self.dropout = torch.nn.Dropout(p=0.5)
        self.readout = LICell(self.hidden_features, self.output_features)

        self.saved_log_probs = []
        self.rewards = []

    def forward(self, x):
        scale = 50
        x_pos = self.constant_current_encoder(torch.nn.functional.relu(scale * x))
        x_neg = self.constant_current_encoder(torch.nn.functional.relu(-scale * x))
        x = torch.cat([x_pos, x_neg], dim=2)

        seq_length, batch_size, _ = x.shape

        # state for hidden layer
        s1 = None
        # state for output layer
        so = None

        voltages = torch.zeros(
            seq_length, batch_size, self.output_features, device=x.device
        )

        # sequential integration loop
        for ts in range(seq_length):
            z1, s1 = self.lif_layer(x[ts, :, :], s1)
            z1 = self.dropout(z1)
            vo, so = self.readout(z1, so)
            voltages[ts, :, :] = vo

        m, _ = torch.max(voltages, 0)
        p_y = torch.nn.functional.softmax(m, dim=1)
        return p_y


def select_action(state, policy, device):
    state = torch.from_numpy(state).float().unsqueeze(0).to(device)
    probs = policy(state)
    m = torch.distributions.Categorical(probs)
    action = m.sample()
    policy.saved_log_probs.append(m.log_prob(action))
    return action.item()


def finish_episode(policy, optimizer):
    eps = np.finfo(np.float32).eps.item()
    R = 0
    policy_loss = []
    returns = []
    for r in policy.rewards[::-1]:
        R = r + FLAGS.gamma * R
        returns.insert(0, R)
    returns = torch.as_tensor(returns)
    returns = (returns - returns.mean()) / (returns.std() + eps)
    for log_prob, R in zip(policy.saved_log_probs, returns):
        policy_loss.append(-log_prob * R)
    optimizer.zero_grad()
    policy_loss = torch.cat(policy_loss).sum()
    policy_loss.backward()
    optimizer.step()
    del policy.rewards[:]
    del policy.saved_log_probs[:]


def main(args):
    running_reward = 10
    torch.manual_seed(FLAGS.random_seed)
    random.seed(FLAGS.random_seed)

    label = f"{FLAGS.policy}-{FLAGS.model}-{FLAGS.random_seed}"
    os.makedirs(f"runs/cartpole/{label}", exist_ok=True)
    os.chdir(f"runs/cartpole/{label}")
    FLAGS.append_flags_into_file("flags.txt")

    np.random.seed(FLAGS.random_seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(FLAGS.random_seed)

    device = torch.device(FLAGS.device)

    env = gym.make(FLAGS.environment)
    env.reset()
    env.seed(FLAGS.random_seed)

    if FLAGS.policy == "ann":
        policy = ANNPolicy().to(device)
    elif FLAGS.policy == "snn":
        policy = Policy().to(device)
    elif FLAGS.policy == "lsnn":
        policy = LSNNPolicy(model=FLAGS.model).to(device)
    optimizer = torch.optim.Adam(policy.parameters(), lr=FLAGS.learning_rate)

    running_rewards = []
    episode_rewards = []

    for e in range(FLAGS.episodes):
        state, ep_reward = env.reset(), 0

        for t in range(1, 10000):  # Don't infinite loop while learning
            action = select_action(state, policy, device=device)
            state, reward, done, _ = env.step(action)
            if FLAGS.render:
                env.render()
            policy.rewards.append(reward)
            ep_reward += reward
            if done:
                break

        running_reward = 0.05 * ep_reward + (1 - 0.05) * running_reward
        finish_episode(policy, optimizer)

        if e % FLAGS.log_interval == 0:
            logging.info(
                "Episode {}/{} \tLast reward: {:.2f}\tAverage reward: {:.2f}".format(
                    e, FLAGS.episodes, ep_reward, running_reward
                )
            )
        episode_rewards.append(ep_reward)
        running_rewards.append(running_reward)
        if running_reward > env.spec.reward_threshold:
            logging.info(
                "Solved! Running reward is now {} and "
                "the last episode runs to {} time steps!".format(running_reward, t)
            )
            break

    np.save("running_rewards.npy", np.array(running_rewards))
    np.save("episode_rewards.npy", np.array(episode_rewards))
    torch.save(optimizer.state_dict(), "optimizer.pt")
    torch.save(policy.state_dict(), "policy.pt")


if __name__ == "__main__":
    app.run(main)



"""
Stateful encoders as torch modules.
"""

from typing import Union, Callable
import torch

from norse.torch.functional import lif
from norse.torch.functional import encode


class ConstantCurrentLIFEncoder(torch.nn.Module):
    def __init__(
        self,
        seq_length: int,
        p: lif.LIFParameters = lif.LIFParameters(),
        dt: float = 0.001,
    ):
        """
        Encodes input currents as fixed (constant) voltage currents, and simulates the spikes that
        occur during a number of timesteps/iterations (seq_length).

        Example:
            >>> data = torch.as_tensor([2, 4, 8, 16])
            >>> seq_length = 2 # Simulate two iterations
            >>> constant_current_lif_encode(data, seq_length)
            (tensor([[0.2000, 0.4000, 0.8000, 0.0000],   # State in terms of membrane voltage
                    [0.3800, 0.7600, 0.0000, 0.0000]]),
            tensor([[0., 0., 0., 1.],                   # Spikes for each iteration
                    [0., 0., 1., 1.]]))

        Parameters:
            seq_length (int): The number of iterations to simulate
            p (LIFParameters): Initial neuron parameters.
            dt (float): Time delta between simulation steps
        """
        super(ConstantCurrentLIFEncoder, self).__init__()
        self.seq_length = seq_length
        self.p = p
        self.dt = dt

    def forward(self, input_currents):
        return encode.constant_current_lif_encode(
            input_currents,
            seq_length=self.seq_length,
            p=self.p,
            dt=self.dt,
        )


class PoissonEncoder(torch.nn.Module):
    def __init__(self, seq_length: int, f_max: float = 100, dt: float = 0.001):
        """
        Encodes a tensor of input values, which are assumed to be in the
        range [0,1] (if not signed, [-1,1] if signed)
        into a tensor of one dimension higher of binary values,
        which represent input spikes.

        Parameters:
            input_values (torch.Tensor): Input data tensor with values assumed to be in the interval [0,1].
            sequence_length (int): Number of time steps in the resulting spike train.
            f_max (float): Maximal frequency (in Hertz) which will be emitted.
            dt (float): Integration time step (should coincide with the integration time step used in the model)
        """
        super(PoissonEncoder, self).__init__()
        self.seq_length = seq_length
        self.f_max = f_max
        self.dt = dt

    def forward(self, x):
        return encode.poisson_encode(x, self.seq_length, f_max=self.f_max, dt=self.dt)


class PopulationEncoder(torch.nn.Module):
    def __init__(
        self,
        out_features: int,
        scale: Union[int, torch.Tensor] = None,
        kernel: Callable[[torch.Tensor], torch.Tensor] = encode.gaussian_rbf,
        distance_function: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = encode.euclidean_distance,
    ):
        """
        Encodes a set of input values into population codes, such that each singular input value is represented by
        a list of numbers (typically calculated by a radial basis kernel), whose length is equal to the out_features.

        Population encoding can be visualised by imagining a number of neurons in a list, whose activity increases
        if a number gets close to its "receptive field".

        .. figure:: https://upload.wikimedia.org/wikipedia/commons/thumb/a/a1/PopulationCode.svg/1920px-PopulationCode.svg.png

            Gaussian curves representing different neuron "receptive fields". Image credit: `Andrew K. Richardson`_.

        .. _Andrew K. Richardson: https://com
        super(PopulationEncoder, self).__init__()mons.wikimedia.org/wiki/File:PopulationCode.svg

        Example:
            >>> data = torch.as_tensor([0, 0.5, 1])
            >>> out_features = 3
            >>> PopulationEncoder(out_features).forward(data)
            tensor([[1.0000, 0.8825, 0.6065],
                    [0.8825, 1.0000, 0.8825],
                    [0.6065, 0.8825, 1.0000]])

        Parameters:
            out_features (int): The number of output *per* input value
            scale (torch.Tensor): The scaling factor for the kernels. Defaults to the maximum value of the input.
                                Can also be set for each individual sample.
            kernel: A function that takes two inputs and returns a tensor. The two inputs represent the center value
                    (which changes for each index in the output tensor) and the actual data value to encode respectively.z
                    Defaults to gaussian radial basis kernel function.
            distance_function: A function that calculates the distance between two numbers. Defaults to euclidean.
        """
        super(PopulationEncoder, self).__init__()
        self.out_features = out_features
        self.scale = scale
        self.kernel = kernel
        self.distance_function = distance_function

    def forward(self, input_tensor):
        return encode.population_encode(
            input_tensor,
            self.out_features,
            self.scale,
            self.kernel,
            self.distance_function,
        )


class SignedPoissonEncoder(torch.nn.Module):
    def __init__(self, seq_length: int, f_max: float = 100, dt: float = 0.001):
        """
        Encodes a tensor of input values, which are assumed to be in the
        range [-1,1] (if not signed, [-1,1] if signed)
        into a tensor of one dimension higher of binary values,
        which represent input spikes.

        Parameters:
            sequence_length (int): Number of time steps in the resulting spike train.
            f_max (float): Maximal frequency (in Hertz) which will be emitted.
            dt (float): Integration time step (should coincide with the integration time step used in the model)
        """
        super(SignedPoissonEncoder, self).__init__()
        self.seq_length = seq_length
        self.f_max = f_max
        self.dt = dt

    def forward(self, x):
        return encode.signed_poisson_encode(
            x, self.seq_length, f_max=self.f_max, dt=self.dt
        )


class SpikeLatencyLIFEncoder(torch.nn.Module):
    """Encodes an input value by the time the first spike occurs.
    Similar to the ConstantCurrentLIFEncoder, but the LIF can be
    thought to have an infinite refractory period.

    Parameters:
        sequence_length (int): Number of time steps in the resulting spike train.
        p (LIFParameters): Parameters of the LIF neuron model.
        dt (float): Integration time step (should coincide with the integration time step used in the model)
    """

    def __init__(self, seq_length, p=lif.LIFParameters(), dt=0.001):
        super(SpikeLatencyLIFEncoder, self).__init__()
        self.seq_length = seq_length
        self.p = p
        self.dt = dt

    def forward(self, input_current):
        return encode.spike_latency_lif_encode(
            input_current, self.seq_length, self.p, self.dt
        )


class SpikeLatencyEncoder(torch.nn.Module):
    """
    For all neurons, remove all but the first spike. This encoding basically measures the time it takes for a
    neuron to spike *first*. Assuming that the inputs are constant, this makes sense in that strong inputs spikes
    fast.

    See `R. Van Rullen & S. J. Thorpe (2001): Rate Coding Versus Temporal Order Coding: What the Retinal Ganglion Cells Tell the Visual Cortex <https://doi.org/10.1162/08997660152002852>`_.

    Spikes are identified by their unique position in the input array.

    Example:
        >>> data = torch.as_tensor([[0, 1, 1], [1, 1, 1]])
        >>> encoder = torch.nn.Sequential(
                        ConstantCurrentLIFEncoder()
                        SpikeLatencyEncoder()
                        )
        >>> encoder(data)
        tensor([[0, 1, 1],
                [1, 0, 0]])
    """

    def forward(self, input_spikes):
        return encode.spike_latency_encode(input_spikes)



import torch

from typing import Optional, Tuple

from ..functional.lif_correlation import (
    LIFCorrelationState,
    LIFCorrelationParameters,
    lif_correlation_step,
)
from ..functional.lif import LIFState
from ..functional.correlation_sensor import CorrelationSensorState


class LIFCorrelation(torch.nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        p: LIFCorrelationParameters = LIFCorrelationParameters(),
        dt: float = 0.001,
    ):
        super(LIFCorrelation, self).__init__()
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.p = p
        self.dt = dt

    def forward(
        self,
        input_tensor: torch.Tensor,
        input_weights: torch.Tensor,
        recurrent_weights: torch.Tensor,
        state: Optional[LIFCorrelationState],
    ) -> Tuple[torch.Tensor, LIFCorrelationState]:
        if state is None:
            hidden_features = self.hidden_size
            input_features = self.input_size
            batch_size = input_tensor.shape[0]
            state = LIFCorrelationState(
                lif_state=LIFState(
                    z=torch.zeros(
                        batch_size,
                        hidden_features,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                    v=self.p.lif_parameters.v_leak.detach(),
                    i=torch.zeros(
                        batch_size,
                        hidden_features,
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                ),
                input_correlation_state=CorrelationSensorState(
                    post_pre=torch.zeros(
                        (batch_size, input_features, hidden_features),
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                    correlation_trace=torch.zeros(
                        (batch_size, input_features, hidden_features),
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ).float(),
                    anti_correlation_trace=torch.zeros(
                        (batch_size, input_features, hidden_features),
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ).float(),
                ),
                recurrent_correlation_state=CorrelationSensorState(
                    correlation_trace=torch.zeros(
                        (batch_size, hidden_features, hidden_features),
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                    anti_correlation_trace=torch.zeros(
                        (batch_size, hidden_features, hidden_features),
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                    post_pre=torch.zeros(
                        (batch_size, hidden_features, hidden_features),
                        device=input_tensor.device,
                        dtype=input_tensor.dtype,
                    ),
                ),
            )
        return lif_correlation_step(
            input_tensor,
            state,
            input_weights,
            recurrent_weights,
            self.p,
            self.dt,
        )



import torch
from norse.torch.functional.stdp_sensor import stdp_sensor_step, STDPSensorState


def test_sensor():
    s = STDPSensorState(
        a_post=torch.zeros(10, 5),
        a_pre=torch.zeros(5, 10),
    )

    z_pre = torch.ones(10)
    z_post = torch.ones(5)

    s = stdp_sensor_step(z_pre, z_post, s)

    assert s.a_post.shape == (10, 5)
    assert s.a_pre.shape == (5, 10)
